#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.06 22:00:00                  #
# ================================================== #

import copy
from uuid import uuid4

from PySide6.QtGui import QAction

from pygpt_net.utils import trans


class Presets:
    def __init__(self, window=None):
        """
        Plugin presets controller

        :param window: Window instance
        """
        self.window = window
        self.config_initialized = False
        self.dialog = False
        self.width = 500
        self.height = 500

    def setup(self):
        """Set up plugin presets"""
        self.window.plugin_presets.setup()  # widget dialog

    def new(self):
        """New preset dialog"""
        uuid = str(uuid4())
        self.window.ui.dialog['create'].id = 'plugin.preset'
        self.window.ui.dialog['create'].input.setText("")
        self.window.ui.dialog['create'].current = uuid
        self.window.ui.dialog['create'].show()

    def create(self, id: str, name: str):
        """
        Create preset

        :param id: preset ID
        :param name: preset name
        """
        if name.strip() == "":
            self.window.ui.status("Preset name cannot be empty!")
            return
        presets = self.get_presets()
        presets[id] = {
            'name': name,
            'enabled': {},
            'config': {},
        }
        self.store(presets)
        self.current_to_preset(id)
        self.window.ui.dialog['create'].close()
        self.update_list()
        self.update_menu()

        self.toggle(id)

    def open(self):
        """Open preset editor"""
        if not self.config_initialized:
            self.setup()
            self.config_initialized = True
        if not self.dialog:
            self.window.ui.dialogs.open(
                'preset.plugins.editor',
                width=self.width,
                height=self.height,
            )
            self.dialog = True

    def close(self):
        """Close plugin presets dialog"""
        if self.dialog:
            self.window.ui.dialogs.close('preset.plugins.editor')
            self.dialog = False

    def toggle_editor(self):
        """Toggle plugin presets dialog"""
        if self.dialog:
            self.close()
        else:
            self.open()

    def rename_by_idx(self, idx: int):
        """
        Rename preset by index

        :param idx: preset index
        """
        id = self.get_id_by_idx(idx)
        self.rename(id)

    def rename(self, id: str):
        """
        Rename preset

        :param id: preset ID
        """
        presets = self.get_presets()
        name = ""
        if id in presets:
            name = presets[id]['name']

        # set dialog and show
        self.window.ui.dialog['rename'].id = 'plugin.preset'
        self.window.ui.dialog['rename'].input.setText(name)
        self.window.ui.dialog['rename'].current = id
        self.window.ui.dialog['rename'].show()

    def update_name(self, id: str, name: str):
        """
        Update preset name

        :param id: preset ID
        :param name: preset name
        """
        if name.strip() == "":
            self.window.ui.status("Preset name cannot be empty!")
            return
        presets = self.get_presets()
        if id in presets:
            presets[id]['name'] = name
        self.store(presets)
        self.window.ui.dialog['rename'].close()
        self.update_list()
        self.update_menu()

    def delete(self, id: str):
        """
        Delete preset

        :param id: preset ID
        """
        presets = self.get_presets()
        if id in presets:
            del presets[id]
        self.store(presets)

        # remove current preset if exists
        current = self.window.core.config.get('preset.plugins')
        if current == id:
            self.window.core.config.set('preset.plugins', '')
            self.window.core.config.save()

        self.update_list()
        self.update_menu()

    def store(self, presets: dict):
        """
        Store presets

        :param presets: presets dict
        """
        self.window.core.plugins.replace_presets(presets)
        self.window.core.plugins.save_presets()

    def get_id_by_idx(self, idx: int) -> str:
        """
        Get preset id by index

        :param idx: preset index
        """
        presets = self.get_presets()
        if len(presets) > idx:
            return list(presets.keys())[idx]

    def select_by_idx(self, idx: int):
        """
        Select preset by index

        :param idx: preset index
        """
        id = self.get_id_by_idx(idx)
        self.toggle(id)
        self.update_menu()

    def delete_by_idx(self, idx: int, force:bool = False):
        """
        Delete preset by index

        :param idx: preset index
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='plugin.presets.delete',
                id=idx,
                msg=trans('confirm.preset.delete'),
            )
            return
        id = self.get_id_by_idx(idx)
        self.delete(id)

    def duplicate(self, id: str):
        """
        Duplicate preset

        :param id: preset ID
        """
        presets = self.get_presets()
        duplicate = copy.deepcopy(presets[id])
        new_id = str(uuid4())
        duplicate['name'] = duplicate['name'] + " - copy"
        presets[new_id] = duplicate
        self.store(presets)
        self.update_list()
        self.update_menu()

    def duplicate_by_idx(self, idx: int):
        """
        Duplicate preset by index

        :param idx: preset index
        """
        id = self.get_id_by_idx(idx)
        self.duplicate(id)

    def reset(self, id: str):
        """
        Reset preset

        :param id: preset ID
        """
        presets = self.get_presets()
        if id in presets:
            presets[id]['enabled'] = {}
            presets[id]['config'] = {}
            self.store(presets)
            if self.get_current_id() == id:
                self.toggle(id)
            self.window.ui.status("Preset cleared: " + presets[id]['name'])

    def reset_by_idx(self, idx: int, force: bool = False):
        """
        Reset preset by index

        :param idx: preset index
        :param force: force reset
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='plugin.presets.reset',
                id=idx,
                msg=trans('confirm.preset.clear'),
            )
            return
        id = self.get_id_by_idx(idx)
        self.reset(id)

    def get_preset(self, id: str) -> dict:
        """
        Get preset by id

        :param id: preset ID
        :return: preset dict
        """
        presets = self.get_presets()
        if id in presets:
            return presets[id]

    def get_presets(self) -> dict:
        """
        Get presets dict

        :return: presets dict
        """
        presets = self.window.core.plugins.get_presets()
        if not isinstance(presets, dict):
            presets = {}
            return presets

        # sort by name
        try:
            return dict(sorted(presets.items(), key=lambda item: item[1]['name']))
        except Exception as e:
            self.window.core.debug.log("Error while sorting presets")
            self.window.core.debug.error(e)
            return presets

    def update_list(self):
        """Reload items"""
        items = self.get_presets()
        self.window.plugin_presets.update_list("preset.plugins.list", items)

    def update_menu(self):
        """Update presets menu"""
        presets = self.get_presets()

        # clear presets
        for id in self.window.ui.menu['plugins_presets']:
            self.window.ui.menu['menu.plugins.presets'].removeAction(self.window.ui.menu['plugins_presets'][id])

        # remove separator if exists
        if len(self.window.ui.menu['menu.plugins.presets'].actions()) > 2:
            self.window.ui.menu['menu.plugins.presets'].actions()[-1].deleteLater()

        # add separator
        if len(presets) > 0:
            self.window.ui.menu['menu.plugins.presets'].addSeparator()

        # add presets
        for id in presets:
            preset = presets[id]
            self.window.ui.menu['plugins_presets'][id] = QAction(preset['name'], self.window, checkable=True)
            self.window.ui.menu['plugins_presets'][id].triggered.connect(
                lambda checked=None,
                       id=id: self.window.controller.plugins.presets.toggle(id))
            self.window.ui.menu['plugins_presets'][id].setMenuRole(QAction.MenuRole.NoRole)
            self.window.ui.menu['menu.plugins.presets'].addAction(self.window.ui.menu['plugins_presets'][id])

        # update current preset
        self.update()

    def update(self):
        """Update presets menu"""
        # clear
        for preset in self.window.ui.menu['plugins_presets']:
            self.window.ui.menu['plugins_presets'][preset].setChecked(False)

        # set checked current preset
        preset = self.window.core.config.get('preset.plugins')
        if preset in self.window.ui.menu['plugins_presets']:
            self.window.ui.menu['plugins_presets'][preset].setChecked(True)

    def toggle(self, id: str):
        """
        Toggle preset

        :param id: preset ID to toggle
        """
        self.window.core.config.set('preset.plugins', id)
        self.window.core.config.save()
        self.update_menu()

        # load preset to current settings
        self.preset_to_current()
        self.window.controller.plugins.reconfigure()  # reconfigure plugins

        # update status
        preset = self.get_preset(id)
        if preset:
            self.window.ui.status("Preset loaded: " + preset['name'])

    def get_current_id(self) -> str:
        """
        Get current preset id

        :return: preset id
        """
        return self.window.core.config.get('preset.plugins')

    def save_current(self):
        """Save current settings as preset"""
        id = self.window.core.config.get('preset.plugins')
        if id:
            preset = self.get_preset(id)
            if preset:
                self.current_to_preset(id)

    def preset_to_current(self):
        """Load settings from preset"""
        id = self.window.core.config.get('preset.plugins')
        if id:
            preset = self.get_preset(id)
            if preset:
                self.window.core.config.set('plugins_enabled', copy.deepcopy(preset['enabled']))
                self.window.core.config.set('plugins', copy.deepcopy(preset['config']))
                self.window.core.config.save()
                self.window.core.plugins.apply_all_options()  # apply restored config

                # if settings dialog opened then reinitialize
                if self.window.controller.plugins.settings.config_dialog:
                    self.window.controller.plugins.settings.init()

    def current_to_preset(self, preset_id: str):
        """
        Store current config to preset

        :param preset_id: preset id
        """
        presets = self.get_presets()
        if preset_id in presets:
            presets[preset_id]['enabled'] = copy.deepcopy(self.window.core.config.get('plugins_enabled'))
            presets[preset_id]['config'] = copy.deepcopy(self.window.core.config.get('plugins'))
            self.store(presets)
        self.update_list()
        self.update_menu()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.12 06:00:00                  #
# ================================================== #

import json
import ssl
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from pygpt_net.plugin.base import BasePlugin
from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem

from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_api"
        self.name = "Command: API calls"
        self.description = "Provides the ability to make external API calls"
        self.order = 100
        self.use_locale = True
        self.init_options()

    def init_options(self):
        """Initialize options"""
        keys = {
            "enabled": "bool",
            "name": "text",
            "instruction": "textarea",
            "get_params": "textarea",
            "post_params": "textarea",
            "post_json": "textarea",
            "headers": "textarea",
            "type": {
                "type": "combo",
                "keys": [
                    {"GET": "GET"},
                    {"POST": "POST"},
                    {"POST_JSON": "POST_JSON"},
                ],
            },
            "endpoint": "textarea",
        }
        items = [
            {
                "enabled": True,
                "name": "search_wiki",
                "instruction": "send API call to Wikipedia to search pages by query",
                "get_params": "query, limit",
                "post_params": "",
                "post_json": "",
                "headers": "",
                "type": "GET",
                "endpoint": 'https://en.wikipedia.org/w/api.php?action=opensearch&limit={limit}&format=json&search={query}',
            },
        ]
        desc = "Add your custom API calls here"
        tooltip = "See the documentation for more details about examples, usage and list of predefined placeholders"
        self.add_option(
            "cmds",
            type="dict",
            value=items,
            label="Your custom API calls",
            description=desc,
            tooltip=tooltip,
            keys=keys,
        )
        self.add_option(
            "disable_ssl",
            type="bool",
            value=False,
            label="Disable SSL verify",
            description="Disables SSL verification when making calls to API",
            tooltip="Disable SSL verify",
        )
        self.add_option(
            "timeout",
            type="int",
            value=5,
            label="Timeout",
            description="Connection timeout (seconds)",
            tooltip="Connection timeout (seconds)",
        )
        self.add_option(
            "user_agent",
            type="text",
            value="Mozilla/5.0",
            label="User agent",
            description="User agent to use when making requests, default: Mozilla/5.0",
            tooltip="User agent to use when making requests",
        )

    def setup(self) -> dict:
        """
        Return available config options

        :return: config options
        """
        return self.options

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for item in self.get_option_value("cmds"):
            if not item["enabled"]:
                continue
            cmd = {
                "cmd": item["name"],
                "instruction": item["instruction"],
                "params": [],
            }
            # GET
            if item["get_params"].strip() != "":
                params = self.extract_params(item["get_params"])
                if len(params) > 0:
                    for param in params:
                        cmd["params"].append(param)
            # POST
            if item["post_params"].strip() != "":
                params = self.extract_params(item["post_params"])
                if len(params) > 0:
                    for param in params:
                        cmd["params"].append(param)

            data['cmd'].append(cmd)

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            for my_cmd in self.get_option_value("cmds"):
                if not my_cmd["enabled"]:
                    continue
                if my_cmd["name"] == item["cmd"]:
                    is_cmd = True
                    my_commands.append(item)

        if not is_cmd:
            return

        try:
            # worker
            worker = Worker()
            worker.plugin = self
            worker.cmds = my_commands
            worker.ctx = ctx

            # signals (base handlers)
            worker.signals.finished.connect(self.handle_finished)
            worker.signals.log.connect(self.handle_log)
            worker.signals.debug.connect(self.handle_debug)
            worker.signals.status.connect(self.handle_status)
            worker.signals.error.connect(self.handle_error)

            # check if async allowed
            if not self.window.core.dispatcher.async_allowed(ctx):
                worker.run()
                return

            # start
            self.window.threadpool.start(worker)

        except Exception as e:
            self.error(e)

    def get_item(self, name: str) -> dict:
        """
        Get API call by name

        :param name: call name
        :return: call dict
        """
        for item in self.get_option_value("cmds"):
            if item["name"] == name:
                return item
        return {}

    def extract_params(self, text: str) -> list:
        """
        Extract params from params string

        :param text: text
        :return: params list
        """
        params = []
        if text is None or text == "":
            return params
        params_list = text.split(",")
        for param in params_list:
            param = param.strip()
            if param == "":
                continue
            params.append({
                "name": param,
                "type": "str",
                "description": param,
            })
        return params

    def call_get(self, url: str, extra_headers: dict = None) -> bytes:
        """
        Call API with GET method

        :param url: URL
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # log API call details
        log_data = {
            "url": url,
            "headers": headers,
            "type": "GET",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data

    def call_post(self, url: str, data: dict, extra_headers: dict = None) -> bytes:
        """
        Call API with POST method

        :param url: URL
        :param data: POST data dict
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # encode data
        data = urlencode(data).encode()

        # log API call details
        log_data = {
            "url": url,
            "data": data,
            "headers": headers,
            "type": "POST",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            data=data,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data

    def call_post_json(self, url: str, data: dict, extra_headers: dict = None) -> bytes:
        """
        Call API with POST method (JSON)

        :param url: URL
        :param data: POST data dict
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
            'Content-Type': 'application/json',
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # encode data
        data = json.dumps(data).encode()

        # log API call details
        log_data = {
            "url": url,
            "data": data,
            "headers": headers,
            "type": "POST JSON",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            data=data,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data

    def log(self, msg: str):
        """
        Log message to console

        :param msg: message to log
        """
        full_msg = '[API] ' + str(msg)
        self.debug(full_msg)
        self.window.ui.status(full_msg)
        if self.is_log():
            print(full_msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.09 21:00:00                  #
# ================================================== #

import os

from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QFileDialog

from pygpt_net.core.dispatcher import Event
from pygpt_net.utils import trans


class Common:
    def __init__(self, window=None):
        """
        Chat common controller

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Set up UI"""
        # stream mode
        if self.window.core.config.get('stream'):
            self.window.ui.nodes['input.stream'].setChecked(True)
        else:
            self.window.ui.nodes['input.stream'].setChecked(False)

        # send clear
        if self.window.core.config.get('send_clear'):
            self.window.ui.nodes['input.send_clear'].setChecked(True)
        else:
            self.window.ui.nodes['input.send_clear'].setChecked(False)

        # send with enter/shift/disabled
        mode = self.window.core.config.get('send_mode')
        if mode == 2:
            self.window.ui.nodes['input.send_shift_enter'].setChecked(True)
            self.window.ui.nodes['input.send_enter'].setChecked(False)
            self.window.ui.nodes['input.send_none'].setChecked(False)
        elif mode == 1:
            self.window.ui.nodes['input.send_enter'].setChecked(True)
            self.window.ui.nodes['input.send_shift_enter'].setChecked(False)
            self.window.ui.nodes['input.send_none'].setChecked(False)
        elif mode == 0:
            self.window.ui.nodes['input.send_enter'].setChecked(False)
            self.window.ui.nodes['input.send_shift_enter'].setChecked(False)
            self.window.ui.nodes['input.send_none'].setChecked(True)

        # cmd enabled
        if self.window.core.config.get('cmd'):
            self.window.ui.nodes['cmd.enabled'].setChecked(True)
        else:
            self.window.ui.nodes['cmd.enabled'].setChecked(False)

        # output timestamps
        self.window.ui.nodes['output.timestamp'].setChecked(self.window.core.config.get('output_timestamp'))

        # raw (plain) output
        self.window.ui.nodes['output.raw'].setChecked(self.window.core.config.get('render.plain'))

        # edit icons
        if self.window.core.config.has('ctx.edit_icons'):
            self.window.ui.nodes['output.edit'].setChecked(self.window.core.config.get('ctx.edit_icons'))

        # images generation
        if self.window.core.config.get('img_raw'):
            self.window.ui.config['global']['img_raw'].setChecked(True)
        else:
            self.window.ui.config['global']['img_raw'].setChecked(False)

        # set focus to input
        self.window.ui.nodes['input'].setFocus()

    def append_to_input(self, text: str, separator: str = "\n"):
        """
        Append text to input

        :param text: text to append
        :param separator: text separator
        """
        prev_text = self.window.ui.nodes['input'].toPlainText()
        cur = self.window.ui.nodes['input'].textCursor()
        cur.movePosition(QTextCursor.End)
        text = str(text).strip()
        if prev_text.strip() != "":
            text = separator + text
        s = text
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertBlock()
        cur.movePosition(QTextCursor.End)  # Move cursor to end of text
        self.window.ui.nodes['input'].setTextCursor(cur)  # Update visible cursor
        self.window.ui.nodes['input'].setFocus()  # Set focus to input

        # update tokens counter
        self.window.controller.ui.update_tokens()

    def clear_input(self):
        """Clear input"""
        self.window.ui.nodes['input'].clear()

    def toggle_stream(self, value: bool):
        """
        Toggle stream

        :param value: value of the checkbox
        """
        self.window.core.config.set('stream', value)

    def toggle_cmd(self, value: bool):
        """
        Toggle cmd enabled

        :param value: value of the checkbox
        """
        self.window.core.config.set('cmd', value)

        # stop commands thread if running
        if not value:
            self.window.controller.command.stop = True
        else:
            self.window.controller.command.stop = False

        self.window.controller.ui.update_tokens()  # update tokens counters

    def toggle_send_clear(self, value: bool):
        """
        Toggle send clear

        :param value: value of the checkbox
        """
        self.window.core.config.set('send_clear', value)

    def toggle_send_shift(self, value: bool):
        """
        Toggle send with shift

        :param value: value of the checkbox
        """
        self.window.core.config.set('send_mode', value)

    def focus_input(self):
        """Focus input"""
        self.window.ui.nodes['input'].setFocus()

    def lock_input(self):
        """Lock input"""
        self.window.controller.chat.input.locked = True
        self.window.ui.nodes['input.send_btn'].setEnabled(False)
        self.window.ui.nodes['input.stop_btn'].setVisible(True)

    def unlock_input(self):
        """Unlock input"""
        self.window.controller.chat.input.locked = False
        self.window.ui.nodes['input.send_btn'].setEnabled(True)
        self.window.ui.nodes['input.stop_btn'].setVisible(False)

    def stop(self, exit: bool = False):
        """
        Stop all

        :param exit: True if app exit
        """
        mode = self.window.core.config.get('mode')
        event = Event(Event.FORCE_STOP, {
            "value": True,
        })
        self.window.core.dispatcher.dispatch(event)  # stop event
        event = Event(Event.AUDIO_INPUT_TOGGLE, {
            "value": False,
        })
        self.window.controller.agent.flow.on_stop()
        self.window.controller.assistant.threads.stop = True
        self.window.controller.assistant.threads.reset()  # reset run and func calls
        self.window.core.dispatcher.dispatch(event)  # stop audio input
        self.window.controller.chat.input.stop = True
        self.window.core.gpt.stop()
        self.unlock_input()

        # remotely stop assistant
        if mode == "assistant" and not exit:
            try:
                self.window.controller.assistant.run_stop()
            except Exception as e:
                self.window.core.debug.log(e)

        self.window.controller.chat.input.generating = False
        self.window.ui.status(trans('status.stopped'))
        self.window.stateChanged.emit(self.window.STATE_IDLE)

    def check_api_key(self) -> bool:
        """
        Check if API KEY is set

        :return: True if API KEY is set, False otherwise
        """
        result = True
        if self.window.core.config.get('api_key') is None or self.window.core.config.get('api_key') == '':
            self.window.controller.launcher.show_api_monit()
            self.window.ui.status("Missing API KEY!")
            result = False
        return result

    def toggle_timestamp(self, value: bool):
        """
        Toggle timestamp

        :param value: value of the checkbox
        """
        self.window.core.config.set('output_timestamp', value)
        self.window.core.config.save()
        self.window.controller.ctx.refresh()

    def toggle_raw(self, value: bool):
        """
        Toggle raw (plain) output

        :param value: value of the checkbox
        """
        self.window.core.config.set('render.plain', value)
        self.window.core.config.save()

        # update checkbox in settings dialog
        self.window.controller.config.checkbox.apply(
            'config',
            'render.plain',
            {'value': value},
        )
        if not value:
            self.window.controller.theme.markdown.update(
                force=True,
            )  # with state store
        else:
            self.window.controller.theme.markdown.clear()

        # restore previous font size
        self.window.controller.ui.update_font_size()

    def toggle_edit_icons(self, value: bool):
        """
        Toggle edit icons

        :param value: value of the checkbox
        """
        self.window.core.config.set('ctx.edit_icons', value)
        self.window.core.config.save()
        self.window.controller.ctx.refresh()

    def img_enable_raw(self):
        """Enable help for images"""
        self.window.core.config.set('img_raw', True)
        self.window.core.config.save()

    def img_disable_raw(self):
        """Disable help for images"""
        self.window.core.config.set('img_raw', False)
        self.window.core.config.save()

    def img_toggle_raw(self, state: bool):
        """
        Toggle help for images

        :param state: state of checkbox
        """
        if not state:
            self.img_disable_raw()
        else:
            self.img_enable_raw()

    def save_text(self, text: str):
        """
        Save text to file

        :param text: text to save
        """
        last_dir = self.window.core.config.get_last_used_dir()
        options = QFileDialog.Options()
        selected_filter = "Text Files (*.txt)"
        file_name, _ = QFileDialog.getSaveFileName(
            self.window,
            "Save as text file",
            last_dir,
            "All Files (*);;Text Files (*.txt);;Python Files (*.py);;Markdown Files (*.md)",
            selected_filter,
            options,
        )
        if file_name:
            # convert text to plain text
            self.window.core.config.set_last_used_dir(
                os.path.dirname(file_name),
            )
            with open(file_name, 'w', encoding="utf-8") as f:
                f.write(
                    str(text).strip()
                )
            self.window.ui.status(trans('status.saved'))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction

from pygpt_net.utils import trans


class Video:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup video menu"""
        self.window.ui.menu['video.capture'] = QAction(
            trans("menu.video.capture"),
            self.window,
            checkable=True,
        )
        self.window.ui.menu['video.capture'].setToolTip(trans('vision.capture.enable.tooltip'))
        self.window.ui.menu['video.capture'].triggered.connect(
            lambda: self.window.controller.camera.toggle(self.window.ui.menu['video.capture'].isChecked())
        )

        self.window.ui.menu['video.capture.auto'] = QAction(
            trans("menu.video.capture.auto"),
            self.window,
            checkable=True,
        )
        self.window.ui.menu['video.capture.auto'].setToolTip(trans('vision.capture.auto.tooltip'))
        self.window.ui.menu['video.capture.auto'].triggered.connect(
            lambda: self.window.controller.camera.toggle_auto(self.window.ui.menu['video.capture.auto'].isChecked())
        )

        """
        self.window.ui.menu['video.player'] = QAction(
            trans("menu.video.player"),
            self.window,
            checkable=True,
        )
        self.window.ui.menu['video.player'].setToolTip(trans('vision.player.tooltip'))
        self.window.ui.menu['video.player'].triggered.connect(
            lambda: self.window.controller.video.show_hide_player(self.window.ui.menu['video.player'].isChecked())
        )
        """

        self.window.ui.menu['menu.video'] = self.window.menuBar().addMenu(trans("menu.video"))
        self.window.ui.menu['menu.video'].addAction(self.window.ui.menu['video.capture'])
        self.window.ui.menu['menu.video'].addAction(self.window.ui.menu['video.capture.auto'])
        self.window.ui.menu['menu.video'].addSeparator()
        # self.window.ui.menu['menu.video'].addAction(self.window.ui.menu['video.player'])
        self.window.ui.menu['menu.video'].setToolTipsVisible(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.16 12:00:00                  #
# ================================================== #

import json
import re
from datetime import datetime

from PySide6.QtCore import Slot

from pygpt_net.plugin.base import BasePlugin
from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem

from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_history"
        self.name = "Context history (calendar, inline)"
        self.type = ["cmd.inline"]
        self.description = "Provides real-time access to context history database"
        self.input_text = None
        self.allowed_cmds = [
            "get_ctx_list_in_date_range",
            "get_ctx_content_by_id",
            "count_ctx_in_date",
            "get_day_note",
            "add_day_note",
            "update_day_note",
            "remove_day_note",
        ]
        self.order = 100
        self.use_locale = True
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.add_option(
            "use_tags",
            type="bool",
            value=False,
            label="Enable: using context @ ID tags",
            description="When enabled, it allows to automatically retrieve context history using @ tags, "
                        "e.g. use @123 in question to retrieve summary of context with ID 123",
        )
        self.add_option(
            "model_summarize",
            type="combo",
            value="gpt-3.5-turbo",
            label="Model",
            description="Model used for summarize, default: gpt-3.5-turbo",
            tooltip="Summarize model",
            use="models",
        )
        self.add_option(
            "summary_max_tokens",
            type="int",
            value=1500,
            label="Max summary tokens",
            description="Max tokens in output when generating summary",
            min=0,
            max=None,
        )
        self.add_option(
            "ctx_items_limit",
            type="int",
            value=30,
            label="Max contexts to retrieve",
            description="Max items in context history list to retrieve in one query. 0 = no limit",
            min=0,
            max=None,
        )
        self.add_option(
            "chunk_size",
            type="int",
            value=100000,
            label="Per-context items content chunk size",
            description="Per-context content chunk size (max characters per chunk)",
            min=1,
            max=None,
        )
        self.add_option(
            "prompt_tag_system",
            type="textarea",
            value="ADDITIONAL CONTEXT: Use the following JSON summary of previous discussions as additional context, "
                  "instead of using commands for retrieve content: {context}",
            label="Prompt: tag_system",
            description="Prompt for use @ tag (system)",
            advanced=True,
        )
        self.add_option(
            "prompt_tag_summary",
            type="textarea",
            value="You are an expert in context summarization. "
                  "Please summarize the given discussion (ID: {id}) by addressing the query and preparing it to serve "
                  "as additional context for a new discussion or to continue the current one: {query}",
            label="Prompt: tag_summary",
            description="Prompt for use @ tag (summary)",
            advanced=True,
        )

        # commands
        self.add_cmd(
            "get_ctx_list_in_date_range",
            instruction="get list of context history (previous conversations between you and me), using date-range "
                        "syntax: \"@date(YYYY-MM-DD)\" for single day, \"@date(YYYY-MM-DD,)\" for date FROM, "
                        "\"@date(,YYYY-MM-DD)\" for date TO, and \"@date(YYYY-MM-DD,YYYY-MM-DD)\" for date FROM-TO",
            params=[
                {
                    "name": "range_query",
                    "type": "str",
                    "description": "range query",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows getting the list of context history (previous conversations)",
        )
        self.add_cmd(
            "get_ctx_content_by_id",
            instruction="get summarized content of context by its ID, use summary query to ask another "
                        "model to summarize content, e.g. \"Summarize following discussion answering the query: (query)\"",
            params=[
                {
                    "name": "id",
                    "type": "int",
                    "description": "context ID",
                    "required": True,
                },
                {
                    "name": "summary_query",
                    "type": "str",
                    "description": "query",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows getting summarized content of context with defined ID",
        )
        self.add_cmd(
            "count_ctx_in_date",
            instruction="count items of context history (previous conversations between us), by providing year, "
                        "month, day or combination of them",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows counting contexts in date range",
        )
        self.add_cmd(
            "get_day_note",
            instruction="get day notes for date",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows retrieving day note for specific date",
        )
        self.add_cmd(
            "add_day_note",
            instruction="add day note",
            params=[
                {
                    "name": "note",
                    "type": "str",
                    "description": "content",
                    "required": True,
                },
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows adding day note for specific date",
        )
        self.add_cmd(
            "update_day_note",
            instruction="update day note",
            params=[
                {
                    "name": "note",
                    "type": "str",
                    "description": "content",
                    "required": True,
                },
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows updating day note for specific date",
        )
        self.add_cmd(
            "remove_day_note",
            instruction="remove day note",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows removing day note for specific date",
        )

    def setup(self) -> dict:
        """
        Return available config options

        :return: config options
        """
        return self.options

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name in [
            Event.CMD_SYNTAX_INLINE,  # inline is allowed
            Event.CMD_SYNTAX,
        ]:
            self.cmd_syntax(data)

        # append current time to system prompt if time plugin is not enabled
        elif name == Event.SYSTEM_PROMPT:
            data['value'] = self.on_system_prompt(data['value'])

        elif name == Event.POST_PROMPT:
            data['value'] = self.on_post_prompt(
                data['value'],
                ctx,
            )

        elif name == Event.USER_SEND:
            self.on_user_send(data['value'])

        elif name in [
            Event.CMD_INLINE,  # inline is allowed
            Event.CMD_EXECUTE,
        ]:
            self.cmd(
                ctx,
                data['commands'],
            )

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        # current time
        if not self.window.controller.plugins.is_type_enabled("time"):
            prompt += "\nCurrent time is: " + datetime.now().strftime('%A, %Y-%m-%d %H:%M:%S')
        return prompt

    def on_post_prompt(self, prompt: str, ctx: CtxItem) -> str:
        """
        Event: POST_PROMPT

        :param prompt: prompt
        :param ctx: context
        :return: updated prompt
        """
        # disable if internal context (autonomous mode)
        if ctx.internal:
            return prompt

        # hashtags
        if self.get_option_value("use_tags") and self.input_text:
            extra = self.handle_tags(self.input_text)
            if extra:
                prompt += "\n" + self.get_option_value("prompt_tag_system").format(
                    context=extra,
                )
        self.input_text = None  # reset input text
        return prompt

    def handle_tags(self, text: str) -> str:
        """
        Handle ID tags in input text

        :param text: input text
        :return: summary of contexts in JSON format
        """
        summary = ""
        contexts = []
        hashtags = self.extract_tags(text)
        ids = []
        for tag in hashtags:
            if tag.isdigit():
                id = int(tag)
                ids.append(id)
        for id in ids:
            try:
                prompt = self.get_option_value("prompt_tag_summary").format(
                    id=id,
                    query=self.input_text,
                )
            except Exception as e:
                self.log("Incorrect prompt: " + str(e))
                return ""
            summary = self.get_summary(id, prompt)
            if summary:
                self.log("@using: " + str(id))
                contexts.append({id: summary})
        if contexts:
            summary = json.dumps(contexts)
        return summary

    def extract_tags(self, text: str) -> list:
        """
        Extract tags from text

        :param text: text
        :return: list of tags
        """
        return re.findall(r"@(\d+)", text)

    def on_user_send(self, text: str):
        """
        Event: USER_SEND

        :param text: text
        """
        self.input_text = text

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if self.has_cmd(option):
                data['cmd'].append(self.get_cmd(option))  # append command

    @Slot()
    def handle_updated(self):
        """Handle updated signal"""
        self.window.controller.calendar.setup()

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Events: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds and self.has_cmd(item["cmd"]):
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        # worker
        worker = Worker()
        worker.plugin = self
        worker.cmds = my_commands
        worker.ctx = ctx

        # signals (base handlers)
        worker.signals.updated.connect(self.handle_updated)
        worker.signals.finished.connect(self.handle_finished)
        worker.signals.log.connect(self.handle_log)
        worker.signals.debug.connect(self.handle_debug)
        worker.signals.status.connect(self.handle_status)
        worker.signals.error.connect(self.handle_error)

        # check if async allowed
        if not self.window.core.dispatcher.async_allowed(ctx):
            worker.run()
            return

        # start
        self.window.threadpool.start(worker)

    def get_day_note(self, year: int, month: int, day: int) -> str:
        """
        Get day note

        :param year: year
        :param month: month
        :param day: day
        :return: day note
        """
        return self.window.core.calendar.load_note(year, month, day)

    def add_day_note(self, year: int, month: int, day: int, note: str) -> bool:
        """
        Add day note

        :param year: year
        :param month: month
        :param day: day
        :param note: note
        :return: True if success
        """
        return self.window.core.calendar.append_to_note(year, month, day, note)

    def update_day_note(self, year: int, month: int, day: int, note: str) -> bool:
        """
        Update day note

        :param year: year
        :param month: month
        :param day: day
        :param note: note
        :return: True if success
        """
        return self.window.core.calendar.update_note(year, month, day, note)

    def remove_day_note(self, year: int, month: int, day: int) -> bool:
        """
        Remove day note

        :param year: year
        :param month: month
        :param day: day
        :return: True if success
        """
        return self.window.core.calendar.remove_note(year, month, day)

    def get_list(self, range: str) -> list:
        """
        Get context list in date range

        :param range: date range
        :return: list of context
        """
        limit = int(self.get_option_value("ctx_items_limit"))
        return self.window.core.ctx.get_list_in_date_range(range, limit=limit)

    def get_summary(self, id: int, prompt: str) -> str:
        """
        Get context content summary by ID

        :param id: context ID
        :param prompt: prompt for summarization
        :return: context content summary
        """
        chunk_size = int(self.get_option_value("chunk_size"))
        items = self.window.core.ctx.get_items_by_id(id)
        if items is None or len(items) == 0:
            return ""
        content = "\n".join(items)
        chunks = self.to_chunks(content, chunk_size)  # it returns list of chunks
        summary = self.get_summarized_text(chunks, prompt)
        return summary

    def count_ctx_in_date(self, year: int = None, month: int = None, day: int = None) -> dict:
        """
        Get context counters

        :param year: year
        :param month: month
        :param day: day
        :return: counters dict
        """
        return self.window.core.ctx.provider.get_ctx_count_by_day(
            year=year,
            month=month,
            day=day,
        )

    def to_chunks(
            self,
            text: str,
            chunk_size: int
    ) -> list:
        """
        Split text into chunks

        :param text: text to split
        :param chunk_size: chunk size
        :return: list of chunks
        """
        if text is None or text == "":
            return []
        return [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]

    def get_summarized_text(
            self,
            chunks: list,
            sys_prompt: str = None
    ) -> str:
        """
        Get summarized text from chunks, TODO: create external method for chunk summarization

        :param chunks: chunks of text
        :param sys_prompt: system prompt
        :return: summarized text
        """
        summary = ""

        # get custom prompt if set
        max_tokens = int(self.get_option_value("summary_max_tokens"))

        # get model
        model = self.window.core.models.from_defaults()
        tmp_model = self.get_option_value("summary_model")
        if self.window.core.models.has(tmp_model):
            model = self.window.core.models.get(tmp_model)

        # summarize per chunk
        for chunk in chunks:
            try:
                response = self.window.core.bridge.quick_call(
                    prompt=chunk,
                    system_prompt=sys_prompt,
                    max_tokens=max_tokens,
                    model=model,
                )
                if response is not None and response != "":
                    summary += response
            except Exception as e:
                self.error(e)
        return summary

    def log(self, msg: str):
        """
        Log message to console

        :param msg: message to log
        """
        full_msg = '[History] ' + str(msg)
        self.debug(full_msg)
        self.window.ui.status(full_msg)
        if self.is_log():
            print(full_msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.15 10:00:00                  #
# ================================================== #

import os
import sys
import traceback
import logging

from pathlib import Path

from pygpt_net.config import Config


class Debug:
    def __init__(self, window=None):
        """
        Debug core

        :param window: Window instance
        """
        self.window = window

    @staticmethod
    def init(level: int = logging.ERROR):
        """
        Initialize error handler

        :param level: log level (default: ERROR)
        """
        if not os.path.exists(os.path.join(Path.home(), '.config', Config.CONFIG_DIR)):
            os.makedirs(os.path.join(Path.home(), '.config', Config.CONFIG_DIR))

        logging.basicConfig(
            level=level,
            format='%(asctime)s - %(levelname)s - %(message)s',
            filename=str(Path(os.path.join(Path.home(), '.config', Config.CONFIG_DIR, 'app.log'))),
            filemode='a',
            encoding='utf-8',
        )

        def handle_exception(exc_type, value, tb):
            """
            Handle uncaught exception

            :param exc_type: exception type
            :param value: exception value
            :param tb: traceback
            """
            logger = logging.getLogger()
            if not hasattr(logging, '_is_handling_exception'):
                logging._is_handling_exception = True
                logger.error("Uncaught exception:", exc_info=(exc_type, value, tb))
                traceback.print_exception(exc_type, value, tb)
                del logging._is_handling_exception  # remove flag when done
            else:
                traceback.print_exception(exc_type, value, tb)

        sys.excepthook = handle_exception

    def switch_log_level(self, level: int):
        """
        Set log level

        :param level: log level
        """
        logging.getLogger().setLevel(level)

    def get_log_level(self) -> int:
        """
        Get log level

        :return: log level
        """
        return logging.getLogger().getEffectiveLevel()

    def get_log_level_name(self) -> str:
        """
        Get current log level name

        :return: log level name
        """
        return self.get_log_level_name_by_id(
            self.get_log_level()
        )

    def get_log_level_name_by_id(self, id: int) -> str:
        """
        Get log level name by id

        :param id: log level id
        :return: log level name
        """
        if id == logging.ERROR:
            return "error"
        elif id == logging.WARNING:
            return "warning"
        elif id == logging.INFO:
            return "info"
        elif id == logging.DEBUG:
            return "debug"
        else:
            return "unknown"

    def info(self, message: any = None, console: bool = True):
        """
        Handle info message

        :param message: message to log
        :param console: print to console

        """
        self.log(
            message,
            logging.INFO,
            console=console,
        )

    def debug(self, message: any = None, console: bool = True):
        """
        Handle debug message

        :param message: message to log
        :param console: print to console
        """
        self.log(
            message,
            logging.DEBUG,
            console=console,
        )

    def warning(self, message: any = None, console: bool = True):
        """
        Handle warning message

        :param message: message to log
        :param console: print to console
        """
        self.log(
            message,
            logging.WARNING,
            console=console,
        )

    def error(self, message: any = None, console: bool = True):
        """
        Handle error message

        :param message: message to log
        :param console: print to console
        """
        self.log(
            message,
            logging.ERROR,
            console=console,
        )

    def log(self, message: any = None, level: int = logging.ERROR, console: bool = True):
        """
        Handle logger message (by level), default level is ERROR

        :param message: message to log
        :param level: logging level (default: ERROR)
        :param console: print to console if True
        """
        if message is None:
            return

        logger = logging.getLogger()

        try:
            # string message
            if isinstance(message, str):
                logger.log(level, message)
                if self.has_level(level) and console:
                    print(message)

            # exception
            elif isinstance(message, Exception):
                is_sys, data = self.parse_exception(message)
                msg = "Exception: {}".format(str(message))
                if not is_sys:
                    msg += "\n{}".format(data)
                logger.log(level, msg, exc_info=is_sys)
                if self.has_level(level) and console and data:
                    print(data)

            # other messages
            else:
                logger.log(level, message)
                if self.has_level(level) and console:
                    print(message)
        except Exception as e:
            pass

        try:
            # send to logger window
            if self.window is not None:
                self.window.logger_message.emit(message)
        except Exception as e:
            pass

    def parse_exception(self, e = None, limit: int = 4) -> (bool, str):
        """
        Parse exception traceback

        :param e: exception
        :param limit: limit of traceback
        :return: sys error, parsed exception as string
        """
        is_sys = False
        type_name = ""
        etype, value, tb = sys.exc_info()  # from sys by default
        if etype is None and e is not None:
            tb = e.__traceback__  # from exception
            type_name = type(e).__name__
            value = str(e)
        else:
            if etype is not None:
                is_sys = True
                type_name = etype.__name__

        # traceback
        traceback_details = traceback.extract_tb(tb)
        if len(traceback_details) >= limit:
            last_calls = traceback_details[-limit:]
        else:
            last_calls = traceback_details
        formatted_traceback = ""
        if last_calls:
            formatted_traceback = "".join(traceback.format_list(last_calls))

        # parse data
        data = ""
        if type_name:
            data += "Type: {}".format(type_name)
        if value:
            data += "Message: {}".format(value)
        if formatted_traceback:
            data += "\nTraceback: {}".format(formatted_traceback)

        return is_sys, data

    def parse_alert(self, msg: any) -> str:
        """
        Parse alert message

        :param msg: message to parse
        :return: parsed message
        """
        if isinstance(msg, Exception):
            is_sys, data = self.parse_exception(msg)
            return "Exception: {}\n{}".format(str(msg), data)
        return str(msg)

    def has_level(self, level: int) -> bool:
        """
        Check if logging level is enabled

        :param level: logging level
        :return: True if enabled
        """
        return level >= logging.getLogger().getEffectiveLevel()

    def enabled(self) -> bool:
        """
        Check if debug is enabled

        :return: True if enabled
        """
        if self.window is not None and self.window.controller.debug.logger_enabled():
            return True
        if self.has_level(logging.DEBUG):
            return True
        return False

    def begin(self, id: str):
        """
        Begin debug data (debug window)

        :param id: debug id
        """
        self.window.controller.dialogs.debug.begin(id)

    def end(self, id: str):
        """
        End debug data (debug window)

        :param id: debug id
        """
        self.window.controller.dialogs.debug.end(id)

    def add(self, id: str, k: str, v: any):
        """
        Append debug entry (debug window)

        :param id: debug id
        :param k: key
        :param v: value
        """
        self.window.controller.dialogs.debug.add(id, k, v)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.12 21:00:00                  #
# ================================================== #

import base64
import json
import os
import re


class Vision:
    def __init__(self, window=None):
        """
        Vision wrapper

        :param window: Window instance
        """
        self.window = window
        self.attachments = {}
        self.urls = []
        self.input_tokens = 0

    def send(self, **kwargs):
        """
        Call OpenAI API for chat with vision

        :param kwargs: keyword arguments
        :return: response or stream chunks
        """
        # get kwargs
        prompt = kwargs.get("prompt", "")
        stream = kwargs.get("stream", False)
        max_tokens = kwargs.get("max_tokens", 200)
        system_prompt = kwargs.get("system_prompt", "")
        attachments = kwargs.get("attachments", {})
        model = kwargs.get("model", None)
        model_id = model.id
        client = self.window.core.gpt.get_client()

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            attachments=attachments,
            model=model,
        )
        response = client.chat.completions.create(
            messages=messages,
            model=model_id,
            max_tokens=int(max_tokens),
            stream=stream,
        )
        return response

    def build(self, **kwargs) -> list:
        """
        Build chat messages list

        :param kwargs: keyword arguments
        :return: messages list
        """
        prompt = kwargs.get("prompt", "")
        system_prompt = kwargs.get("system_prompt", None)
        attachments = kwargs.get("attachments", {})
        model = kwargs.get("model", None)
        messages = []

        # tokens config
        mode = 'vision'
        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_tokens = self.window.core.config.get('max_total_tokens')

        # fit to max model tokens
        if max_tokens > model.ctx:
            max_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # append initial (system) message
        if system_prompt is not None and system_prompt != "":
            messages.append({"role": "system", "content": system_prompt})
        else:
            if system_prompt is not None and system_prompt != "":
                messages.append({"role": "system", "content": system_prompt})

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_prompt_items(
                model.id,
                mode,
                used_tokens,
                max_tokens,
            )
            for item in items:
                # input
                if item.input is not None and item.input != "":
                    content = self.build_content(item.input)
                    messages.append({"role": "user", "content": content})

                # output
                if item.output is not None and item.output != "":
                    content = self.build_content(item.output)
                    messages.append({"role": "assistant", "content": content})

        # append current prompt
        content = self.build_content(prompt, attachments)
        messages.append({"role": "user", "content": content})

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def build_content(self, prompt: str, attachments: dict = None) -> list:
        """
        Build vision contents

        :param prompt: prompt (user input)
        :param attachments: attachments (dict, optional)
        :return: List of contents
        """
        content = [
            {
                "type": "text",
                "text": str(prompt)
            }
        ]

        self.attachments = {}  # reset attachments, only current prompt
        self.urls = []

        # extract URLs from prompt
        urls = self.extract_urls(prompt)
        if len(urls) > 0:
            for url in urls:
                content.append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": url,
                        }
                    }
                )
                self.urls.append(url)

        # local images (attachments)
        if attachments is not None and len(attachments) > 0:
            for id in attachments:
                attachment = attachments[id]
                if os.path.exists(attachment.path):
                    # check if it's an image
                    if self.is_image(attachment.path):
                        base64_image = self.encode_image(attachment.path)
                        content.append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{base64_image}",
                                }
                            }
                        )
                        self.attachments[id] = attachment.path

        return content

    def extract_urls(self, text: str) -> list:
        """
        Extract img urls from text

        :param text: text
        :return: list of img urls
        """
        urls = re.findall(r'(https?://\S+)', text)
        img_urls = []
        for url in urls:
            if self.is_image(url):
                img_urls.append(url)

        # if single url, return it
        if "".join(urls).strip() == text.strip():
            return urls

        return img_urls

    def is_image(self, path: str) -> bool:
        """
        Check if url is image

        :param path: url
        :return: True if image
        """
        return path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.webp'))

    def encode_image(self, image_path: str) -> str:
        """
        Encode image to base64

        :param image_path: path to image
        :return: base64 encoded image
        """
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_attachments(self) -> dict:
        """
        Get attachments

        :return: attachments dict
        """
        return self.attachments

    def get_urls(self) -> list:
        """
        Get urls

        :return: urls list
        """
        return self.urls

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.15 10:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QTextEdit, QMenu

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class CalendarNote(QTextEdit):
    def __init__(self, window=None):
        """
        Notepad

        :param window: main window
        """
        super(CalendarNote, self).__init__(window)
        self.window = window
        self.setAcceptRichText(False)
        self.setStyleSheet(self.window.controller.theme.style('font.chat.output'))
        self.value = self.window.core.config.data['font_size']
        self.textChanged.connect(self.window.controller.calendar.note.update)
        self.max_font_size = 42
        self.min_font_size = 8

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        selected_text = self.textCursor().selectedText()
        if selected_text:
            # plain text
            plain_text = self.textCursor().selection().toPlainText()

            # audio read
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            # copy to
            copy_to_menu = QMenu(trans('text.context_menu.copy_to'), self)

            # input
            action = QAction(QIcon(":/icons/input.svg"), trans('text.context_menu.copy_to.input'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.append_to_input(selected_text))
            copy_to_menu.addAction(action)

            # notepad
            num_notepads = self.window.controller.notepad.get_num_notepads()
            if num_notepads > 0:
                for i in range(1, num_notepads + 1):
                    name = self.window.controller.notepad.get_notepad_name(i)
                    action = QAction(QIcon(":/icons/paste.svg"), name, self)
                    action.triggered.connect(lambda checked=False, i=i:
                                             self.window.controller.notepad.append_text(selected_text, i))
                    copy_to_menu.addAction(action)

            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text))
            menu.addAction(action)
        else:
            # save as (all)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText()))
            menu.addAction(action)

        menu.exec_(event.globalPos())

    def audio_read_selection(self):
        """
        Read selected text (audio)
        """
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            self.window.core.config.data['font_size'] = self.value
            self.window.core.config.save()
            option = self.window.controller.settings.editor.get_option('font_size')
            option['value'] = self.value
            self.window.controller.config.apply(
                parent_id='config',
                key='font_size',
                option=option,
            )
            self.window.controller.ui.update_font_size()
            event.accept()
        else:
            super(CalendarNote, self).wheelEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.17 13:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import Event


class Prompt:
    def __init__(self, window=None):
        """
        Prompt core

        :param window: Window instance
        """
        self.window = window

    def get(self, prompt: str) -> str:
        """
        Get prompt content

        :param prompt: id of the prompt
        :return: text content
        """
        key = "prompt." + prompt
        if self.window.core.config.has(key):
            return str(self.window.core.config.get(key))
        return ""

    def build_final_system_prompt(self, prompt: str) -> str:
        # tmp dispatch event: system prompt
        event = Event(Event.SYSTEM_PROMPT, {
            'mode': self.window.core.config.get('mode'),
            'value': prompt,
            'silent': True,
        })
        self.window.core.dispatcher.dispatch(event)
        prompt = event.data['value']

        if self.window.core.config.get('cmd') or self.window.controller.plugins.is_type_enabled("cmd.inline"):
            
            # cmd syntax tokens
            data = {
                'prompt': prompt,
                'silent': True,
                'syntax': [],
                'cmd': [],
            }

            # tmp dispatch event: command syntax apply
            # full execute cmd syntax
            if self.window.core.config.get('cmd'):
                event = Event(Event.CMD_SYNTAX, data)
                self.window.core.dispatcher.dispatch(event)
                prompt = self.window.core.command.append_syntax(event.data)

            # inline cmd syntax only
            elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
                event = Event(Event.CMD_SYNTAX_INLINE, data)
                self.window.core.dispatcher.dispatch(event)
                prompt = self.window.core.command.append_syntax(event.data)

        return prompt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.27 15:00:00                  #
# ================================================== #

import json
import time


class NotepadItem:
    def __init__(self):
        self.id = 0
        self.uuid = None
        self.idx = 0
        self.title = ""
        self.content = ""
        self.deleted = False
        ts = int(time.time())
        self.created = ts
        self.updated = ts
        self.initialized = False

    def to_dict(self):
        return {
            'id': self.id,
            'uuid': self.uuid,
            'idx': self.idx,
            'title': self.title,
            'content': self.content,
            'deleted': self.deleted,
            'created': self.created,
            'updated': self.updated,
            'initialized': self.initialized
        }

    def dump(self):
        """
        Dump item to string

        :return: serialized item
        :rtype: str
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

import os

from PySide6.QtGui import QTextCursor

from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Interpreter:
    def __init__(self, window=None):
        """
        Python real-time interpreter controller

        :param window: Window instance
        """
        self.window = window
        self.opened = False
        self.is_edit = False
        self.auto_clear = True

        # interpreter data files in /data directory
        self.file_current = ".interpreter.current.py"
        self.file_input = ".interpreter.input.py"
        self.file_output = ".interpreter.output.py"

    def setup(self):
        """Setup"""
        self.load_output()
        self.update()

    def update(self):
        """Update icon"""
        pass
        """
        self.toggle_icon(self.window.controller.plugins.is_type_enabled('interpreter'))
        """

    def append_output(self, output: str, type="stdout", **kwargs):
        """
        Append output to the interpreter window

        :param output: Output data
        :param type: Output type
        :param kwargs: Additional parameters
        """
        if self.is_edit:
            self.disable_edit()

        if type == "stdin":
            data = ">> "  + str(output)
        else:
            data = str(output)
        cur = self.window.interpreter.textCursor()
        cur.movePosition(QTextCursor.End)
        s = data + "\n"
        while s:
            head, sep, s = s.partition("\n")
            cur.insertText(head)
            if sep:  # New line if LF
                cur.insertText("\n")
        self.window.interpreter.setTextCursor(cur)
        self.save_output()

    def load_input(self):
        """Load input data from file"""
        data = self.load_prev_input()
        self.window.interpreter.setPlainText(data)
        self.window.interpreter.setFocus()

    def load_output(self):
        """Load output data from file"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        content = ""
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                content = f.read()
        self.window.interpreter.setPlainText(content)
        self.window.ui.nodes['interpreter.input'].setFocus()

    def save_output(self):
        """Save output data to file"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.window.interpreter.toPlainText())

    def load_prev_input(self) -> str:
        """
        Load previous input data from file

        :return: Input data
        """
        input = ""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                input = f.read()
        return input

    def save_input(self, input: str):
        """
        Save input data

        :param input: Input data
        """
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        with open(path , "w", encoding="utf-8") as f:
            f.write(input)

    def clear_input(self):
        """Clear input"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        if os.path.exists(path):
            os.remove(path)
        self.window.interpreter.clear()

    def clear_output(self):
        """Clear output"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        if os.path.exists(path):
            os.remove(path)
        self.window.interpreter.clear()

    def clear(self, force: bool = False):
        """
        Clear current window

        :param force: Force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='interpreter.clear',
                id=0,
                msg=trans("interpreter.clear.confirm"),
            )
            return
        if self.is_edit:
            self.clear_input()
        else:
            self.clear_output()

    def clear_all(self):
        """Clear input and output"""
        self.clear_output()
        self.clear_input()

    def send_input(self):
        """Send input to interpreter"""
        # switch to output mode if edit mode is enabled
        if self.is_edit:
            self.save_edit()
            self.disable_edit()

        if self.auto_clear:
            self.clear_output()

        input = str(self.window.ui.nodes['interpreter.input'].toPlainText())
        if self.is_all():
            cmd = "code_execute_all"
        else:
            if input == "":
                return
            cmd = "code_execute"

        commands = [
            {
                "cmd": cmd,
                "params": {
                    "code": input,
                    "path": self.file_current,
                },
                "silent": True,
            }
        ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
        })
        event.ctx = CtxItem()  # tmp
        self.window.controller.command.dispatch_only(event)
        self.window.ui.nodes['interpreter.input'].clear()
        self.window.ui.nodes['interpreter.input'].setFocus()

    def toggle_edit(self):
        """Toggle edit mode"""
        self.is_edit = self.window.ui.nodes['interpreter.edit'].isChecked()
        self.window.interpreter.setReadOnly(not self.is_edit)

        if self.is_edit:
            self.window.ui.nodes['interpreter.edit_label'].setText(trans("interpreter.edit_label.edit"))
            self.load_input()
        else:
            self.window.ui.nodes['interpreter.edit_label'].setText(trans("interpreter.edit_label.output"))
            self.load_output()

        self.cursor_to_end()

    def save_edit(self):
        """Save edit data to file"""
        if not self.is_edit:
            return
        data = self.window.interpreter.toPlainText()
        self.save_input(data)

    def open(self):
        """Open interpreter dialog"""
        self.opened = True
        self.window.ui.dialogs.open('interpreter', width=800, height=600)
        self.window.ui.nodes['interpreter.input'].setFocus()
        self.cursor_to_end()

    def close(self):
        """Close interpreter dialog"""
        self.opened = False
        self.window.ui.dialogs.close('interpreter')

    def toggle(self):
        """Toggle interpreter dialog open/close"""
        if self.opened:
            self.close()
        else:
            self.open()

    def toggle_icon(self, state: bool):
        """
        Toggle interpreter icon

        :param state: State
        """
        self.window.ui.nodes['icon.interpreter'].setVisible(state)

    def toggle_auto_clear(self):
        """Toggle auto clear"""
        self.auto_clear = self.window.ui.nodes['interpreter.auto_clear'].isChecked()

    def get_data(self) -> str:
        """
        Get current data from interpreter

        :return: Output data
        """
        return self.window.interpreter.toPlainText()

    def is_all(self) -> bool:
        """
        Check if execute all is enabled

        :return: True if execute all is enabled
        """
        return self.window.ui.nodes['interpreter.all'].isChecked()

    def cursor_to_end(self):
        """Move cursor to end"""
        cur = self.window.interpreter.textCursor()
        cur.movePosition(QTextCursor.End)
        self.window.interpreter.setTextCursor(cur)

    def disable_edit(self):
        """Disable edit mode"""
        self.window.ui.nodes['interpreter.edit'].setChecked(False)
        self.toggle_edit()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QTextEdit, QApplication, QPlainTextEdit
import pygpt_net.icons_rc


class PythonInput(QTextEdit):
    def __init__(self, window=None):
        """
        Python interpreter input

        :param window: main window
        """
        super(PythonInput, self).__init__(window)
        self.window = window
        self.setAcceptRichText(False)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'interpreter-input')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.setFocus()

    def update_stylesheet(self, data: str):
        """
        Update stylesheet

        :param data: stylesheet CSS
        """
        self.setStyleSheet(self.default_stylesheet + data)

    def audio_read_selection(self):
        """Read selected text (audio)"""
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        super(PythonInput, self).keyPressEvent(event)
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            mode = self.window.core.config.get('send_mode')
            if mode > 0:  # Enter or Shift + Enter
                if mode == 2:  # Shift + Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers == QtCore.Qt.ShiftModifier:
                        self.window.controller.interpreter.send_input()
                else:  # Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers != QtCore.Qt.ShiftModifier:
                        self.window.controller.interpreter.send_input()
                self.setFocus()

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            size_str = f"{self.value}px"
            self.update_stylesheet(f"font-size: {size_str};")
            event.accept()
        else:
            super(PythonInput, self).wheelEvent(event)


class PythonOutput(QPlainTextEdit):
    def __init__(self, window=None):
        """
        Python interpreter output

        :param window: main window
        """
        super(PythonOutput, self).__init__(window)
        self.window = window
        self.setReadOnly(True)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.textChanged.connect(self.window.controller.interpreter.save_edit)
        self.setProperty('class', 'interpreter-output')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)

    def update_stylesheet(self, data: str):
        """
        Update stylesheet

        :param data: stylesheet CSS
        """
        self.setStyleSheet(self.default_stylesheet + data)

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            size_str = f"{self.value}px"
            self.update_stylesheet(f"font-size: {size_str};")
            event.accept()
        else:
            super(PythonOutput, self).wheelEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.06 02:00:00                  #
# ================================================== #

import json

from sqlalchemy import text


class Viewer:
    def __init__(self, database):
        self.database = database
        self.tables = {}
        self.auto_backup = True

    def fetch_data(
            self,
            table: str,
            columns: list,
            sort_by: str,
            order: str,
            search_query: str = None,
            search_column: str = None,
            offset: int = 0,
            limit: int = 100
    ) -> list:
        """
        Fetch data from the database with the given parameters.

        :param table: Table name
        :param columns: List of columns to fetch
        :param sort_by: Column to sort by
        :param order: Sort order (ASC or DESC)
        :param search_query: Search query
        :param search_column: Search column
        :param offset: Offset
        :param limit: Limit
        :return: Fetched data
        """
        tables = self.database.get_tables()

        if search_column:
            search_fields = [search_column]
        else:
            search_fields = tables[table]['search_fields']

        base_query = f"SELECT {', '.join(columns)} FROM {table}"
        where_clause = ""
        order_clause = f" ORDER BY {sort_by} {order}"
        limit_clause = f" LIMIT {limit} OFFSET {offset}"

        params = {}
        if search_query:
            search_clauses = [f"{column} LIKE :search_query" for column in search_fields]
            where_clause = f" WHERE {' OR '.join(search_clauses)}"
            params['search_query'] = f"%{search_query}%"

        query = f"{base_query}{where_clause}{order_clause}{limit_clause}"
        stmt = text(query).bindparams(**params)
        with self.database.get_db().connect() as conn:
            result = conn.execute(stmt).fetchall()
            return result

    def count_rows(
            self,
            table: str,
            search_query: str = None,
            search_column: str = None
    ) -> int:
        """
        Count the number of rows in the table with an optional search query.

        :param table: Table name
        :param search_query: Search query
        :param search_column: Search column
        :return: Number of rows
        """
        base_query = f"SELECT COUNT(*) FROM {table}"
        where_clause = ""
        params = {}
        tables = self.database.get_tables()

        if search_column:
            search_fields = [search_column]
        else:
            search_fields = tables[table]['search_fields']

        if search_query:
            where_clause = f" WHERE {' OR '.join([f'{column} LIKE :search_query' for column in search_fields])}"
            params['search_query'] = f"%{search_query}%"

        query = f"{base_query}{where_clause}"
        stmt = text(query).bindparams(**params)
        with self.database.get_db().connect() as conn:
            count = conn.execute(stmt).scalar()
            return count

    def is_auto_backup(self) -> bool:
        """
        Get auto backup status.

        :return: Auto backup status
        """
        return self.database.window.ui.debug["db"].browser.is_auto_backup()

    def delete_row(self, data: dict):
        """
        Delete row from the database.

        :param data: Dictionary with table and row_id keys
        """
        # create backup
        if self.is_auto_backup():
            backup_path = self.database.make_backup()
            if backup_path:
                msg = f"[DB] Created DB backup: {backup_path}"
                self.log(msg)

        # delete row
        with self.database.get_db().begin() as conn:
            conn.execute(
                text(f"DELETE FROM {data['table']} WHERE id = :row_id")
                .bindparams(row_id=data['row_id'])
            )
            msg = f"[DB] Deleted row ID {data['row_id']} from table {data['table']}"
            self.log(msg)
        self.database.window.ui.debug["db"].browser.update_table_view()

    def update_row(self, data: dict):
        """
        Update row in the database.

        :param data: Dictionary with table, id, field and value keys
        """
        field = data['field']
        value = data['value']
        tables = self.database.get_tables()
        json_columns = tables[data['table']]['json_columns']
        timestamp_columns = tables[data['table']]['timestamp_columns']
        primary_key = tables[data['table']]['primary_key']

        # check JSON
        if field in json_columns or field.endswith("_json"):
            try:
                value = json.dumps(json.loads(value))  # validate and pack JSON
            except:
                raise ValueError(f"Invalid JSON value for column {field}")

        # check timestamp
        if field in timestamp_columns or field.endswith("_ts"):
            try:
                value = int(value)
            except:
                raise ValueError(f"Invalid timestamp value for column {field}")

        # check foreign id field
        if field.endswith("_id"):
            try:
                value = int(value)
            except:
                raise ValueError(f"Invalid _id value for column {field}")

        # create backup
        if self.is_auto_backup():
            backup_path = self.database.make_backup()
            if backup_path:
                msg = f"[DB] Created DB backup: {backup_path}"
                self.log(msg)

        # update row
        with self.database.get_db().begin() as conn:
            conn.execute(
                text(f"UPDATE {data['table']} SET {data['field']} = :value WHERE {primary_key} = :id")
                .bindparams(id=data['id'], value=value)
            )
            msg = f"[DB] Updated row ID {data['id']} in table {data['table']}"
            self.log(msg)
        self.database.window.ui.debug["db"].browser.update_table_view()

    def truncate_table(self, data: dict, reset: bool = False):
        """
        Truncate table in the database.

        :param data: Dictionary with table key
        :param reset: Reset table sequence
        """
        # create backup
        if self.is_auto_backup():
            backup_path = self.database.make_backup()
            if backup_path:
                msg = f"[DB] Created DB backup: {backup_path}"
                self.log(msg)

        # truncate table
        with self.database.get_db().begin() as conn:
            conn.execute(text(f"DELETE FROM {data['table']}"))
            if reset:  # reset table sequence (autoincrement)
                conn.execute(text(f"DELETE FROM sqlite_sequence WHERE name='{data['table']}'"))
                msg = f"[DB] Truncated table {data['table']}"
            else:
                msg = f"[DB] Deleted all rows from table {data['table']}"
            self.log(msg)
        self.database.window.ui.debug["db"].browser.update_table_view()

    def log(self, msg: str):
        """
        Log action to debug.

        :param msg: Message
        """
        self.database.window.core.debug.info(msg)
        print(msg)
        self.database.window.ui.status(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.23 06:00:00                  #
# ================================================== #

import os
import shutil
import sys

from packaging.version import parse as parse_version, Version


class Patch:
    def __init__(self, window=None, provider=None):
        self.window = window
        self.provider = provider

    def execute(self, version: Version) -> bool:
        """
        Migrate to current app version

        :param version: current app version
        :return: True if migrated
        """
        # return
        # if old version is < 2.0.162 and if json file exists
        path = os.path.join(self.window.core.config.path, 'indexes.json')
        if os.path.exists(path):
            self.provider.truncate_all()
            self.import_from_json()
            os.rename(path, path + ".old")  # rename to "indexes.json.old"
            print("BACKUP: indexes.json file has been renamed to indexes.json.old")
            return True

    def import_from_json(self) -> bool:
        """
        Import idx data from JSON

        :return: True if imported
        """
        # use old json provider to load old indexes
        provider = self.window.core.idx.providers['json_file']
        provider.attach(self.window)

        print("[DB] Migrating idx into database storage...")
        print("[DB] Importing old idx data from JSON files... this may take a while. Please wait...")

        stores = provider.load()
        cols, _ = shutil.get_terminal_size()
        all = 0
        for store_id in stores:
            store = stores[store_id]
            c = len(store)
            i = 0
            for idx_id in store:
                idx = store[idx_id]
                line = "[DB] %s - importing idx %s/%s: %s" % (store_id, i + 1, c, idx.name)
                print(f"{line:<{cols}}", end='\r')
                sys.stdout.flush()
                self.import_idx(
                    store_id,
                    idx_id,
                    idx.items,
                )
                i += 1
                all += 1
                print()  # new line

        if all > 0:
            print("[DB] [DONE] Imported %s indexes." % all)
            return True

    def import_idx(self, store_id: str, idx_id: str, items: list) -> bool:
        """
        Import index data from JSON

        :param store_id: store ID
        :param idx_id: idx ID
        :param items: list of items
        """
        for file_id in items:
            item = items[file_id]
            self.provider.append_file(store_id, idx_id, item)
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

import json

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QApplication

from pygpt_net.core.dispatcher import Event
from pygpt_net.core.worker import Worker, WorkerSignals


class Command:
    def __init__(self, window=None):
        """
        Commands dispatch controller

        :param window: Window instance
        """
        self.window = window
        self.stop = False

    def dispatch(self, event: Event, all: bool = False):
        """
        Dispatch cmd execute event (command execution)

        :param event: event object
        :param all: True to dispatch to all plugins
        """
        self.window.core.debug.info("Dispatch CMD event begin: " + event.name)
        if self.window.core.debug.enabled():
            self.window.core.debug.debug("EVENT BEFORE: " + str(event))

        for id in self.window.core.plugins.get_ids():
            if self.window.controller.plugins.is_enabled(id) or all:
                if event.stop or (event.name == Event.CMD_EXECUTE and self.is_stop()):
                    if self.is_stop():
                        self.stop = False  # unlock needed here
                    break
                if self.window.core.debug.enabled():
                    self.window.core.debug.debug("Apply [{}] to plugin: ".format(event.name) + id)

                self.window.stateChanged.emit(self.window.STATE_BUSY)
                self.window.core.dispatcher.apply(id, event)

    def dispatch_only(self, event: Event):
        """
        Dispatch cmd execute event only (command execution)

        :param event: event object
        """
        self.window.core.debug.info("Dispatch CMD event begin: " + event.name)
        for id in self.window.core.plugins.get_ids():
            self.window.core.dispatcher.apply(id, event)

    def dispatch_async(self, event: Event):
        """
        Dispatch async cmd event (command execution)

        :param event: event object
        """
        worker = Worker(self.worker)
        worker.signals = WorkerSignals()
        worker.signals.finished.connect(self.handle_finished)
        worker.kwargs['event'] = event
        worker.kwargs['window'] = self.window
        worker.kwargs['finished_signal'] = worker.signals.finished
        self.window.threadpool.start(worker)

    def worker(self, event: Event, window, finished_signal: Signal):
        """
        Command worker callback

        :param event: event object
        :param window: Window instance
        :param finished_signal: WorkerSignals: finished signal
        """
        for id in window.core.plugins.get_ids():
            if window.controller.plugins.is_enabled(id):
                if event.stop or (event.name == Event.CMD_EXECUTE and self.is_stop()):
                    if self.is_stop():
                        self.stop = False  # unlock needed here
                    break
                window.core.dispatcher.apply(id, event, is_async=True)
        finished_signal.emit(event)

    def is_stop(self) -> bool:
        """
        Check if stop is requested

        :return: True if stop is requested
        """
        return self.stop

    def handle_debug(self, data: any):
        """
        Handle thread debug log

        :param data to log
        """
        self.window.core.debug.info(data)

    def handle_finished(self, event: Event):
        """
        Handle command execution finish (response from sync execution)

        :param event: event object
        """
        ctx = event.ctx
        if ctx is not None:
            self.window.core.debug.info("Reply...")
            if self.window.core.debug.enabled():
                self.window.core.debug.debug("CTX REPLY: " + str(ctx))

            self.window.ui.status("")  # Clear status
            if ctx.reply:
                data = json.dumps(ctx.results)
                if ctx.extra_ctx:
                    data = ctx.extra_ctx  # if extra content is set, use it as data to send
                prev_ctx = self.window.core.ctx.as_previous(ctx)  # copy result to previous ctx and clear current ctx
                self.window.controller.chat.input.send(
                    text=data,
                    force=True,
                    internal=ctx.internal,
                    prev_ctx=prev_ctx,
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.15 10:00:00                  #
# ================================================== #

import os
import shutil

from PySide6 import QtGui, QtCore
from PySide6.QtWidgets import QFileDialog, QApplication

from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.dispatcher import Event
from pygpt_net.utils import trans


class Image:
    def __init__(self, window=None):
        """
        Image controller

        :param window: Window instance
        """
        self.window = window

    def send(
            self,
            text: str,
            prev_ctx: CtxItem = None
    ) -> CtxItem:
        """
        Send prompt for image generate

        :param text: prompt for image generation
        :param prev_ctx: previous ctx item
        :return: ctx item
        """
        num = int(self.window.ui.config['global']['img_variants'].input.text() or 1)
        if num < 1:
            num = 1
        elif num > 4:
            num = 4

        # force 1 image if dall-e-3 model is used
        model = self.window.core.config.get('model')
        model_data = self.window.core.models.get(model)
        if model_data.id == 'dall-e-3':
            num = 1

        self.window.ui.status(trans('status.sending'))

        # create ctx item
        ctx = CtxItem()
        ctx.set_input(text, self.window.core.config.get('user_name'))
        ctx.prev_ctx = prev_ctx  # store previous context item

        # event: context before
        event = Event(Event.CTX_BEFORE)
        event.ctx = ctx
        self.window.core.dispatcher.dispatch(event)

        # add ctx to DB
        self.window.core.ctx.add(ctx)
        self.window.controller.chat.render.append_input(ctx)

        # process events to update UI
        QApplication.processEvents()

        # handle ctx name (generate title from summary if not initialized)
        if self.window.core.config.get('ctx.auto_summary'):
            self.window.controller.ctx.prepare_name(ctx)

        # generate image
        try:
            self.window.core.bridge.call(
                mode="image",
                prompt=text,
                ctx=ctx,
                model=model_data,  # model instance
                num=num,
            )
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            self.window.ui.status(trans('status.error'))

        return ctx

    def handle_response(self, ctx: CtxItem, paths: list, prompt: str):
        """
        Handle response

        :param ctx: ctx item
        :param paths: list with paths to downloaded images
        :param prompt: prompt used to generate images
        """
        string = ""
        i = 1
        for path in paths:
            string += "{}) `{}`".format(i, path) + "\n"
            i += 1

        if self.window.core.config.get('img_dialog_open'):
            self.open_images(paths)

        if not self.window.core.config.get('img_raw'):
            string += "\nPrompt: "
            string += prompt

        local_urls = self.window.core.filesystem.make_local_list(paths)
        ctx.images = local_urls  # save images paths
        ctx.set_output(string.strip())

        # event: after context
        event = Event(Event.CTX_AFTER)
        event.ctx = ctx
        self.window.core.dispatcher.dispatch(event)

        # store last mode (in text mode this is handled in send_text)
        mode = self.window.core.config.get('mode')
        self.window.core.ctx.post_update(mode)  # post update context, store last mode, etc.
        self.window.controller.chat.render.append_output(ctx)
        self.window.core.ctx.store()  # save current ctx to DB
        self.window.ui.status(trans('status.img.generated'))

        # update ctx in DB
        self.window.core.ctx.update_item(ctx)

        # append extra output to chat
        self.window.controller.chat.render.append_extra(ctx)
        self.window.controller.chat.render.end_extra()

        self.window.stateChanged.emit(self.window.STATE_IDLE)  # set state to idle

    def handle_response_inline(self, ctx: CtxItem, paths: list, prompt: str):
        """
        Handle inline response

        :param ctx: ctx item
        :param paths: list with paths to downloaded images
        :param prompt: prompt used to generate images
        """
        string = ""
        i = 1
        for path in paths:
            string += "{}) `{}`".format(i, path) + "\n"
            i += 1

        local_urls = self.window.core.filesystem.make_local_list(paths)
        ctx.images = local_urls  # save images paths in ctx item here
        self.window.core.ctx.update_item(ctx)  # update in DB
        self.window.ui.status(trans('status.img.generated'))  # update status

        # WARNING:
        # if internal (sync) mode, then re-send OK status response, if not, append only img result
        # it will only inform system that image was generated, user will see it in chat with image after render
        # of ctx item (link to images are appended to ctx item)
        if ctx.internal:
            ctx.results.append(
                {
                    "request": {
                        "cmd": "image",
                    },
                    "result": "OK. Generated {} image(s).".format(len(paths)),
                    "paths": paths,
                }
            )
            ctx.reply = True
            self.window.controller.chat.render.append_extra(ctx)  # show image first
            self.window.controller.chat.render.end_extra()
            self.window.core.dispatcher.reply(ctx)
            return

        # NOT internal-mode, user called, so append only img output to chat (show images now)
        self.window.controller.chat.render.append_extra(ctx)
        self.window.controller.chat.render.end_extra()

        self.window.stateChanged.emit(self.window.STATE_IDLE)  # set state to idle

    def open_images(self, paths: list):
        """
        Open image in dialog

        :param paths: paths to images
        """
        num_images = len(paths)
        resize_to = 512
        if num_images > 1:
            resize_to = 256

        i = 0
        for path in paths:
            pixmap = QtGui.QPixmap(path)
            pixmap = pixmap.scaled(resize_to, resize_to, QtCore.Qt.KeepAspectRatio)
            self.window.ui.nodes['dialog.image.pixmap'][i].path = path
            self.window.ui.nodes['dialog.image.pixmap'][i].setPixmap(pixmap)
            self.window.ui.nodes['dialog.image.pixmap'][i].setVisible(True)
            i += 1

        # hide unused images
        for j in range(i, 4):
            self.window.ui.nodes['dialog.image.pixmap'][j].setVisible(False)

        # resize dialog
        self.window.ui.dialog['image'].resize(520, 520)
        self.window.ui.dialog['image'].show()

    def open(self, path: str):
        """
        Open image in default image viewer

        :param path: path to image
        """
        if os.path.exists(path):
            self.window.controller.files.open(path)

    def open_dir(self, path: str):
        """
        Open image in default image viewer

        :param path: path to image
        """
        if os.path.exists(path):
            self.window.controller.files.open_dir(
                path,
                True,
            )

    def save(self, path: str):
        """
        Save image

        :param path: path to image
        """
        save_path = QFileDialog.getSaveFileName(
            self.window,
            trans('img.save.title'),
            os.path.basename(path),
            "PNG (*.png)",
        )
        if save_path:
            try:
                shutil.copyfile(path, save_path[0])
                self.window.ui.status(trans('status.img.saved'))
            except Exception as e:
                self.window.core.debug.log(e)

    def delete(self, path: str, force: bool = False):
        """
        Delete image

        :param path: path to image
        :param force: force delete without confirmation
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='img_delete',
                id=path,
                msg=trans('confirm.img.delete'),
            )
            return
        try:
            os.remove(path)
            for i in range(0, 4):
                if self.window.ui.nodes['dialog.image.pixmap'][i].path == path:
                    self.window.ui.nodes['dialog.image.pixmap'][i].setVisible(False)
        except Exception as e:
            self.window.core.debug.log(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

import datetime
import os
from pathlib import Path

from PySide6.QtCore import Qt, QModelIndex, QDir
from PySide6.QtGui import QAction, QIcon, QCursor, QResizeEvent
from PySide6.QtWidgets import QTreeView, QMenu, QWidget, QVBoxLayout, QFileSystemModel, QLabel, QHBoxLayout, \
    QPushButton, QSizePolicy

from pygpt_net.ui.widget.element.button import ContextMenuButton
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class FileExplorer(QWidget):
    def __init__(self, window, directory, index_data):
        """
        File explorer widget

        :param window: Window instance
        :param directory: directory to explore
        :param index_data: index data
        """
        super().__init__()

        self.window = window
        self.index_data = index_data
        self.directory = directory
        self.model = IndexedFileSystemModel(self.window, self.index_data)
        self.model.setRootPath(directory)
        self.model.setFilter(self.model.filter() | QDir.Hidden)  # show hidden files

        self.treeView = QTreeView()
        self.treeView.setModel(self.model)
        self.treeView.setRootIndex(self.model.index(directory))

        header = QHBoxLayout()

        self.btn_open = QPushButton(trans('action.open'))
        self.btn_open.setMaximumHeight(40)
        self.btn_open.clicked.connect(
                lambda: self.action_open(self.directory)
        )
        self.btn_open.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_upload = QPushButton(trans('files.local.upload'))
        self.btn_upload.setMaximumHeight(40)
        self.btn_upload.clicked.connect(self.window.controller.files.upload_local)
        self.btn_upload.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_idx = ContextMenuButton(trans('idx.btn.index_all'), self)
        self.btn_idx.action = self.idx_context_menu
        self.btn_idx.setMaximumHeight(40)
        self.btn_idx.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_clear = ContextMenuButton(trans('idx.btn.clear'), self)
        self.btn_clear.action = self.clear_context_menu
        self.btn_clear.setMaximumHeight(40)
        self.btn_clear.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.path_label = QLabel(self.directory)
        self.path_label.setMaximumHeight(40)
        self.path_label.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)

        header.addWidget(self.btn_open)
        header.addWidget(self.btn_upload)
        header.addStretch()
        header.addWidget(self.path_label)
        header.addStretch()

        header.addWidget(self.btn_idx)
        header.addWidget(self.btn_clear)
        layout = QVBoxLayout()

        self.window.ui.nodes['tip.output.tab.files'] = HelpLabel(trans('tip.output.tab.files'), self.window)

        layout.addWidget(self.treeView)
        layout.addWidget(self.window.ui.nodes['tip.output.tab.files'])
        layout.addLayout(header)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.treeView.setContextMenuPolicy(Qt.CustomContextMenu)
        self.treeView.customContextMenuRequested.connect(self.openContextMenu)
        self.treeView.setColumnWidth(0, self.width() / 2)

        self.header = self.treeView.header()
        self.header.setStretchLastSection(True)
        self.header.setContentsMargins(0, 0, 0, 0)

        self.column_proportion = 0.3
        self.adjustColumnWidths()

        self.header.setStyleSheet("""
           QHeaderView::section {
               text-align: center;
               vertical-align: middle;
           }
       """)

    def idx_context_menu(self, parent, pos):
        """
        Index all btn context menu

        :param parent: parent widget
        :param pos: mouse  position
        """
        menu = QMenu(self)
        idx_list = self.window.core.config.get('llama.idx.list')
        if len(idx_list) > 0:
            for idx in idx_list:
                id = idx['id']
                name = idx['name'] + " (" + idx['id'] + ")"
                action = menu.addAction("IDX: " + name)
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.index_all_files(id)
                )
        menu.exec_(parent.mapToGlobal(pos))

    def clear_context_menu(self, parent, pos):
        """
        Clear btn context menu

        :param parent: parent widget
        :param pos: mouse position
        """
        menu = QMenu(self)
        idx_list = self.window.core.config.get('llama.idx.list')
        if len(idx_list) > 0:
            for idx in idx_list:
                id = idx['id']
                name = idx['name'] + " (" + idx['id'] + ")"
                action = menu.addAction("IDX: " + name)
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.clear(id)
                )
        menu.exec_(parent.mapToGlobal(pos))

    def adjustColumnWidths(self):
        """Adjust column widths"""
        total_width = self.treeView.width()
        first_column_width = int(total_width * self.column_proportion)
        self.treeView.setColumnWidth(0, first_column_width)
        for column in range(1, self.model.columnCount()):
            self.treeView.setColumnWidth(column, (total_width - first_column_width) // (self.model.columnCount() - 1))

    def resizeEvent(self, event: QResizeEvent):
        """
        Resize event

        :param event: Event object
        """
        super().resizeEvent(event)
        self.adjustColumnWidths()

    def openContextMenu(self, position):
        """
        Open context menu

        :param position: mouse position
        """
        indexes = self.treeView.selectedIndexes()
        if indexes:
            index = indexes[0]
            path = self.model.filePath(index)
            actions = {}

            """
            # play
            if self.window.core.filesystem.is_video(path):
                actions['play_video'] = QAction(
                    QIcon(":/icons/video.svg"),
                    trans('action.video.play'),
                    self,
                )
                actions['play_video'].triggered.connect(
                    lambda: self.window.controller.video.play(path),
                )
            """

            # open file
            actions['open'] = QAction(QIcon(":/icons/view.svg"), trans('action.open'), self)
            actions['open'].triggered.connect(
                lambda: self.action_open(path),
            )

            # open in file manager
            actions['open_dir'] = QAction(QIcon(":/icons/folder_filled.svg"), trans('action.open_dir'), self)
            actions['open_dir'].triggered.connect(
                lambda: self.action_open_dir(path),
            )

            # download
            actions['download'] = QAction(QIcon(":/icons/download.svg"), trans('action.download'), self)
            actions['download'].triggered.connect(
                lambda: self.window.controller.files.download_local(path),
            )

            # rename
            actions['rename'] = QAction(QIcon(":/icons/edit.svg"), trans('action.rename'), self)
            actions['rename'].triggered.connect(
                lambda: self.action_rename(path),
            )

            # duplicate
            actions['duplicate'] = QAction(QIcon(":/icons/stack.svg"), trans('action.duplicate'), self)
            actions['duplicate'].triggered.connect(
                lambda: self.window.controller.files.duplicate_local(path, ""),
            )

            if os.path.isdir(path):
                parent = path
            else:
                parent = os.path.dirname(path)

            # touch file
            actions['touch'] = QAction(QIcon(":/icons/add.svg"), trans('action.touch'), self)
            actions['touch'].triggered.connect(
                lambda: self.window.controller.files.touch_file(parent),
            )

            # make dir in dir
            actions['mkdir'] = QAction(QIcon(":/icons/add_folder.svg"), trans('action.mkdir'), self)
            actions['mkdir'].triggered.connect(
                lambda: self.action_make_dir(parent),
            )

            # refresh
            actions['refresh'] = QAction(QIcon(":/icons/reload.svg"), trans('action.refresh'), self)
            actions['refresh'].triggered.connect(
                lambda: self.window.controller.files.update_explorer(),
            )

            # upload to dir
            actions['upload'] = QAction(QIcon(":/icons/upload.svg"), trans('action.upload'), self)
            actions['upload'].triggered.connect(
                lambda: self.window.controller.files.upload_local(parent),
            )

            # delete
            actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.delete'), self)
            actions['delete'].triggered.connect(
                lambda: self.action_delete(path),
            )

            menu = QMenu(self)
            """
            if "play_video" in actions:
                menu.addAction(actions['play_video'])
            """
            menu.addAction(actions['open'])
            menu.addAction(actions['open_dir'])

            # use
            use_menu = QMenu(trans('action.use'), self)

            # use as attachment
            if not os.path.isdir(path):
                actions['use_attachment'] = QAction(
                    QIcon(":/icons/attachment.svg"),
                    trans('action.use.attachment'),
                    self,
                )
                actions['use_attachment'].triggered.connect(
                    lambda: self.window.controller.files.use_attachment(path),
                )

                # by type
                if self.window.core.filesystem.is_image(path):
                    actions['use_as_image'] = QAction(
                        QIcon(":/icons/brush.svg"),
                        trans('action.use.image'),
                        self,
                    )
                    actions['use_as_image'].triggered.connect(
                        lambda: self.window.controller.painter.open_external(path),
                    )

            # copy work path
            actions['use_copy_work_path'] = QAction(
                QIcon(":/icons/copy.svg"),
                trans('action.use.copy_work_path'),
                self,
            )
            actions['use_copy_work_path'].triggered.connect(
                lambda: self.window.controller.files.copy_work_path(path),
            )

            # copy sys path
            actions['use_copy_sys_path'] = QAction(
                QIcon(":/icons/copy.svg"),
                trans('action.use.copy_sys_path'),
                self,
            )
            actions['use_copy_sys_path'].triggered.connect(
                lambda: self.window.controller.files.copy_sys_path(path),
            )

            # use read cmd
            actions['use_read_cmd'] = QAction(QIcon(":/icons/view.svg"), trans('action.use.read_cmd'), self)
            actions['use_read_cmd'].triggered.connect(
                lambda: self.window.controller.files.make_read_cmd(path),
            )

            # add actions to menu
            if not os.path.isdir(path):
                use_menu.addAction(actions['use_attachment'])
            if "use_as_image" in actions:
                use_menu.addAction(actions['use_as_image'])
            use_menu.addAction(actions['use_copy_work_path'])
            use_menu.addAction(actions['use_copy_sys_path'])
            use_menu.addAction(actions['use_read_cmd'])
            menu.addMenu(use_menu)

            # remove from index
            file_id = self.window.core.idx.files.get_id(path)
            remove_actions = []
            for idx in self.index_data:
                items = self.index_data[idx]
                if file_id in items:
                    action = QAction(QIcon(":/icons/delete.svg"), trans("action.idx.remove") + ": " + idx, self)
                    action.triggered.connect(
                        lambda checked=False,
                               idx=idx,
                               file_id=file_id: self.action_idx_remove(file_id, idx)  # by file_id, not path
                    )
                    remove_actions.append(action)

            # add to index if allowed
            if self.window.core.idx.indexing.is_allowed(path):
                idx_menu = QMenu(trans('action.idx'), self)
                idx_list = self.window.core.config.get('llama.idx.list')
                if len(idx_list) > 0 or len(remove_actions) > 0:
                    # add
                    if len(idx_list) > 0:
                        for idx in idx_list:
                            id = idx['id']
                            name = idx['name'] + " (" + idx['id'] + ")"
                            action = QAction(QIcon(":/icons/search.svg"), "IDX: " + name, self)
                            action.triggered.connect(
                                lambda checked=False,
                                       id=id,
                                       path=path: self.action_idx(path, id)
                            )
                            idx_menu.addAction(action)

                # remove
                if len(remove_actions) > 0:
                    idx_menu.addSeparator()
                    for action in remove_actions:
                        idx_menu.addAction(action)

                menu.addMenu(idx_menu)

            menu.addAction(actions['download'])
            menu.addAction(actions['touch'])
            menu.addAction(actions['mkdir'])
            menu.addAction(actions['upload'])
            menu.addAction(actions['refresh'])

            menu.addAction(actions['rename'])
            menu.addAction(actions['duplicate'])
            menu.addAction(actions['delete'])

            menu.exec(QCursor.pos())
        else:
            # no item selected
            actions = {}

            # touch file
            actions['touch'] = QAction(QIcon(":/icons/add.svg"), trans('action.touch'), self)
            actions['touch'].triggered.connect(
                lambda: self.window.controller.files.touch_file(self.directory),
            )

            # open in file manager
            actions['open_dir'] = QAction(QIcon(":/icons/folder_filled.svg"), trans('action.open_dir'), self)
            actions['open_dir'].triggered.connect(
                lambda: self.action_open_dir(self.directory),
            )

            # make dir in dir
            actions['mkdir'] = QAction(QIcon(":/icons/add_folder.svg"), trans('action.mkdir'), self)
            actions['mkdir'].triggered.connect(
                lambda: self.action_make_dir(self.directory),
            )

            # upload in dir
            actions['upload'] = QAction(QIcon(":/icons/upload.svg"), trans('action.upload'), self)
            actions['upload'].triggered.connect(
                lambda: self.window.controller.files.upload_local(),
            )

            menu = QMenu(self)
            menu.addAction(actions['touch'])
            menu.addAction(actions['open_dir'])
            menu.addAction(actions['mkdir'])
            menu.addAction(actions['upload'])
            menu.exec(QCursor.pos())

    def action_open(self, path):
        """
        Open action handler

        :param path: path to open
        """
        self.window.controller.files.open(path)

    def action_idx(self, path: str, idx: str):
        """
        Index file or dir handler

        :param path: path to open
        :param idx: index ID to use (name)
        """
        self.window.controller.idx.indexer.index_file(path, idx)

    def action_idx_remove(self, path: str, idx: str):
        """
        Remove file or dir from index handler

        :param path: path to open
        :param idx: index ID to use (name)
        """
        self.window.controller.idx.indexer.index_file_remove(path, idx)

    def action_open_dir(self, path: str):
        """
        Open in directory action handler

        :param path: path to open
        """
        self.window.controller.files.open_dir(path, True)

    def action_make_dir(self, path: str = None):
        """
        Make directory action handler

        :param path: parent path
        """
        self.window.controller.files.make_dir_dialog(path)

    def action_rename(self, path: str):
        """
        Rename action handler

        :param path: path to rename
        """
        self.window.controller.files.rename(path)

    def action_delete(self, path: str):
        """
        Delete action handler

        :param path: path to delete
        """
        self.window.controller.files.delete(path)


class IndexedFileSystemModel(QFileSystemModel):
    def __init__(self, window, index_dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window = window
        self.index_dict = index_dict
        self.directoryLoaded.connect(self.refresh_path)

    def refresh_path(self, path):
        index = self.index(path)
        if index.isValid():
            self.dataChanged.emit(index, index)

    def columnCount(self, parent=QModelIndex()) -> int:
        """
        Return column count

        :param parent: parent
        :return: column count
        """
        return super().columnCount(parent) + 1

    def data(self, index, role=Qt.DisplayRole) -> any:
        """
        Data handler

        :param index: row index
        :param role: role
        :return: data
        """
        # index status
        if index.column() == self.columnCount() - 1:
            if role == Qt.DisplayRole:
                file_path = self.filePath(index.siblingAtColumn(0))
                status = self.get_index_status(file_path)   # get index status
                if status['indexed']:
                    ts = datetime.datetime.fromtimestamp(status['last_index_at'])
                    # check if today
                    if ts.date() == datetime.date.today():
                        dt = ts.strftime("%H:%M")
                    else:
                        dt = ts.strftime("%Y-%m-%d %H:%M")
                    content = ''
                    content += dt
                    content += ' (' + ",".join(status['indexed_in']) + ')'
                else:
                    content = '-'  # if file not indexed
                return content
        # modified date
        elif index.column() == self.columnCount() - 2:
            if role == Qt.DisplayRole:
                dt = self.lastModified(index)
                data = dt.toString("yyyy-MM-dd HH:mm:ss")
                ts = dt.toSecsSinceEpoch()
                dt_from_ts = datetime.datetime.fromtimestamp(ts)
                if dt_from_ts.date() == datetime.date.today():
                    data = dt_from_ts.strftime("%H:%M")
                else:
                    data = dt_from_ts.strftime("%Y-%m-%d %H:%M")
                file_path = self.filePath(index.siblingAtColumn(0))
                status = self.get_index_status(file_path)  # get index status
                if status['indexed']:
                    # if modified date is newer, mark file with *
                    if 'last_index_at' in status and status['last_index_at'] < dt.toSecsSinceEpoch():
                        data += '*'
                return data

        return super().data(index, role)

    def get_index_status(self, file_path) -> dict:
        """
        Get index status

        :param file_path: file path
        :return: file index status
        """
        file_id = self.window.core.idx.files.get_id(file_path)
        indexed_in = []
        indexed_timestamps = {}
        last_index_at = 0
        for idx in self.index_dict:
            items = self.index_dict[idx]
            if file_id in items:
                indexed_in.append(idx)  # append idx where file is indexed
                indexed_timestamps[idx] = items[file_id]['indexed_ts']
                if items[file_id]['indexed_ts'] > last_index_at:
                    last_index_at = items[file_id]['indexed_ts']

        # sort indexed_in by timestamp DESC
        indexed_in = sorted(
            indexed_in,
            key=lambda x: indexed_timestamps[x],
            reverse=True,
        )
        if len(indexed_in) > 0:
            return {
                'indexed': True,
                'indexed_in': indexed_in,
                'last_index_at': last_index_at,
            }
        else:
            return {
                'indexed': False,
            }

    def headerData(self, section, orientation, role=Qt.DisplayRole) -> str:
        """
        Prepare Header data (append Indexed column)

        :param section: Section
        :param orientation: Orientation
        :param role: Role
        :return: Header data
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:  # name
                return trans('files.explorer.header.name')
            elif section == 1:  # size
                return trans('files.explorer.header.size')
            elif section == 2:  # type
                return trans('files.explorer.header.type')
            elif section == 3:  # modified
                return trans('files.explorer.header.modified')
            elif section == 4:  # indexed
                return trans('files.explorer.header.indexed')
        return super().headerData(section, orientation, role)

    def update_idx_status(self, idx_data):
        """
        Update index data status

        :param idx_data: new index data dict
        """
        self.index_dict = idx_data
        top_left_index = self.index(0, 0)
        bottom_right_index = self.index(self.rowCount() - 1, self.columnCount() - 1)
        path = self.rootPath()
        self.dataChanged.emit(top_left_index, bottom_right_index, [Qt.DisplayRole])
        self.setRootPath("")
        self.setRootPath(path)
        self.layoutChanged.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QUrl, QTimer
from PySide6.QtMultimedia import QMediaPlayer
from PySide6.QtMultimediaWidgets import QVideoWidget
from PySide6.QtWidgets import (QWidget, QPushButton, QHBoxLayout, QVBoxLayout,
                               QFileDialog, QSlider, QLabel, QStyle, QSizePolicy)

class VideoPlayerWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.mediaPlayer = QMediaPlayer()
        self.loaded = False

        self.update_timer = QTimer(self)
        self.update_timer.setInterval(200)
        self.update_timer.timeout.connect(self.update_ui)

        self.videoWidget = QVideoWidget()
        self.videoWidget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.playPauseButton = QPushButton()
        self.playPauseButton.setEnabled(False)
        self.update_play_pause_icon()

        self.stopButton = QPushButton()
        self.stopButton.setIcon(self.style().standardIcon(QStyle.SP_MediaStop))
        self.stopButton.setEnabled(False)

        self.openButton = QPushButton("Open Video...")

        self.labelDuration = QLabel()
        self.labelCurrentTime = QLabel()

        self.slider = QSlider(Qt.Horizontal)
        self.slider.setRange(0, 0)

        # layout
        controlLayout = QHBoxLayout()
        controlLayout.addWidget(self.openButton)
        controlLayout.addWidget(self.playPauseButton)
        controlLayout.addWidget(self.stopButton)

        sliderLayout = QHBoxLayout()
        sliderLayout.addWidget(self.labelCurrentTime)
        sliderLayout.addWidget(self.slider)
        sliderLayout.addWidget(self.labelDuration)

        layout = QVBoxLayout()
        layout.addWidget(self.videoWidget)
        layout.addLayout(sliderLayout)
        layout.addLayout(controlLayout)
        self.setLayout(layout)

        self.bind_signals()

    def bind_signals(self):
        """Bind signals"""
        self.openButton.clicked.connect(self.open_file)
        self.playPauseButton.clicked.connect(self.toggle_play_pause)
        self.stopButton.clicked.connect(self.stop_video)
        self.slider.sliderMoved.connect(self.set_position)
        # self.videoWidget.mousePressEvent = self.video_widget_clicked

        # connect signals
        self.mediaPlayer.positionChanged.connect(self.position_changed)
        self.mediaPlayer.durationChanged.connect(self.duration_changed)
        self.mediaPlayer.playbackStateChanged.connect(self.on_playback_state_changed)

        self.mediaPlayer.setVideoOutput(self.videoWidget)
        self.mediaPlayer.mediaStatusChanged.connect(self.media_status_changed)

    def media_status_changed(self, status):
        """
        Media status changed

        :param status: status
        """
        if status == QMediaPlayer.LoadedMedia and not self.loaded:
            self.loaded = True
            self.play_video()

    def open_file(self):
        """Open file"""
        self.loaded = False
        path, _ = QFileDialog.getOpenFileName(self, "Open Video")
        if path:
            self.open(path)

    def open(self, path: str):
        """
        Open video

        :param path: video path
        """
        self.reset()
        self.mediaPlayer.setSource(QUrl.fromLocalFile(path))
        self.playPauseButton.setEnabled(True)
        self.stopButton.setEnabled(True)

    def play_video(self):
        """Play video"""
        self.mediaPlayer.play()
        self.on_playback_state_changed()  # update icon when play/pause is toggled.

    def reset(self):
        """Reset"""
        self.loaded = False
        self.playPauseButton.setEnabled(False)
        self.stopButton.setEnabled(False)
        self.update_timer.stop()
        self.mediaPlayer.stop()
        self.mediaPlayer.setSource(QUrl())
        self.slider.setValue(0)
        self.labelCurrentTime.setText(self.format_time(0))
        self.labelDuration.setText(self.format_time(0))

    def toggle_play_pause(self):
        """Toggle play/pause"""
        if self.mediaPlayer.playbackState() == QMediaPlayer.PlayingState:
            self.mediaPlayer.pause()
        else:
            self.mediaPlayer.play()
        self.on_playback_state_changed()  # update icon when play/pause is toggled.

    def stop_video(self):
        """Stop video"""
        if self.mediaPlayer.playbackState() == QMediaPlayer.PlayingState or self.mediaPlayer.playbackState() == QMediaPlayer.PausedState:
            self.mediaPlayer.stop()

        self.slider.setValue(0)
        self.labelCurrentTime.setText(self.format_time(0))
        self.update_play_pause_icon()

    def on_playback_state_changed(self):
        """Playback state changed"""
        self.update_play_pause_icon()

    def update_play_pause_icon(self):
        """Update play/pause icon"""
        if self.mediaPlayer.playbackState() == QMediaPlayer.PlayingState:
            self.playPauseButton.setIcon(self.style().standardIcon(QStyle.SP_MediaPause))
            self.update_timer.start()
        else:
            self.playPauseButton.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            self.update_timer.stop()

    def video_widget_clicked(self, event):
        """
        Video widget clicked

        :param event: event
        """
        self.toggle_play_pause()

    def position_changed(self, position):
        """
        Position changed

        :param position: position
        """
        self.slider.setValue(position)
        self.labelCurrentTime.setText(self.format_time(position))

    def duration_changed(self, duration):
        """
        Duration changed

        :param duration: duration
        """
        self.slider.setRange(0, duration)
        self.labelDuration.setText(self.format_time(duration))

    def set_position(self, position):
        """
        Set position

        :param position: position
        """
        self.mediaPlayer.setPosition(position)

    def update_ui(self):
        """Update UI"""
        # sync slider with the video time
        self.position_changed(self.mediaPlayer.position())

    def format_time(self, ms):
        """
        Format time

        :param ms: milliseconds
        """
        # convert milliseconds to hours:minutes:seconds format
        seconds = (ms // 1000) % 60
        minutes = (ms // 60000) % 60
        hours = (ms // 3600000)
        return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.18 03:00:00                  #
# ================================================== #

import os.path
import subprocess
import docker

from pygpt_net.item.ctx import CtxItem


class Runner:
    def __init__(self, plugin=None):
        """
        Cmd Runner

        :param plugin: plugin
        """
        self.plugin = plugin
        self.signals = None

    def send_interpreter_input(self, data: str):
        """
        Send input to subprocess

        :param data: input text
        """
        if self.signals is not None:
            self.signals.output.emit(data, "stdin")

    def send_interpreter_output(self, data: str, type: str):
        """
        Send output to interpreter

        :param data: output text
        :param type: output type (stdout/stderr)
        """
        if self.signals is not None:
            self.signals.output.emit(data, type)

    def handle_result(self, stdout, stderr):
        """
        Handle result from subprocess

        :param stdout: stdout
        :param stderr: stderr
        :return: result
        """
        result = None
        if stdout:
            result = stdout.decode("utf-8")
            self.send_interpreter_output(result, "stdout")
            self.log("STDOUT: {}".format(result))
        if stderr:
            result = stderr.decode("utf-8")
            self.send_interpreter_output(result, "stderr")
            self.log("STDERR: {}".format(result))
        if result is None:
            result = "No result (STDOUT/STDERR empty)"
            self.log(result)
        return result

    def handle_result_docker(self, response) -> str:
        """
        Handle result from docker container

        :param response: response
        :return: result
        """
        result = None
        if response:
            result = response.decode('utf-8')
        self.send_interpreter_output(result, "stdout")
        self.log(
            "Result: {}".format(result),
            sandbox=True,
        )
        return result

    def is_sandbox(self) -> bool:
        """
        Check if sandbox is enabled

        :return: True if sandbox is enabled
        """
        return self.plugin.get_option_value('sandbox_docker')

    def get_docker(self) -> docker.client.DockerClient:
        """
        Get docker client

        :return: docker client instance
        """
        return docker.from_env()

    def get_docker_image(self) -> str:
        """
        Get docker image name

        :return: docker image
        """
        return self.plugin.get_option_value('sandbox_docker_image')

    def get_volumes(self) -> dict:
        """
        Get docker volumes

        :return: docker volumes
        """
        path = self.plugin.window.core.config.get_user_dir('data')
        mapping = {}
        mapping[path] = {
            "bind": "/data",
            "mode": "rw",
        }
        return mapping

    def run_docker(self, cmd: str) -> bytes or None:
        """
        Run docker container with command and return response

        :param cmd: command to run
        :return: response
        """
        client = self.get_docker()
        mapping = self.get_volumes()
        response = None
        try:
            response = client.containers.run(
                self.get_docker_image(),
                cmd,
                volumes=mapping,
                working_dir="/data",
                stdout=True,
                stderr=True,
            )
        except Exception as e:
            self.error(e)
            response = str(e).encode("utf-8")
        return response


    def code_execute_file_host(self, ctx, item: dict, request: dict) -> dict or None:
        """
        Execute code from file on host machine

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Executing Python file: {}".format(item["params"]['path'])
        self.log(msg)
        path = self.prepare_path(item["params"]['path'], on_host=True)

        # check if file exists
        if not os.path.isfile(path):
            return {
                "request": request,
                "result": "File not found",
            }

        """
        # send input to interpreter
        with open(path, 'r', encoding="utf-8") as file:
            code = file.read()
            self.append_input(code)
            self.send_interpreter_input(code)  # send input to interpreter
        """

        # run code
        cmd = self.plugin.get_option_value('python_cmd_tpl').format(filename=path)
        self.log("Running command: {}".format(cmd))
        try:
            process = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
        except Exception as e:
            self.error(e)
            stdout = None
            stderr = str(e).encode("utf-8")
        result = self.handle_result(stdout, stderr)
        return {
            "request": request,
            "result": result,
            "context": "PYTHON OUTPUT:\n--------------------------------\n" + result,
        }

    def code_execute_file_sandbox(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute code from file in sandbox (docker)

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        path = item["params"]['path']
        msg = "Executing Python file: {}".format(path)
        self.log(msg, sandbox=True)
        cmd = self.plugin.get_option_value('python_cmd_tpl').format(
            filename=path,
        )

        """
        # send input to interpreter
        with open(self.prepare_path(path, on_host=True), 'r', encoding="utf-8") as file:
            code = file.read()
            self.append_input(code)
            self.send_interpreter_input(code)  # send input to interpreter
        """

        self.log("Running command: {}".format(cmd), sandbox=True)
        response = self.run_docker(cmd)
        result = self.handle_result_docker(response)
        return {
            "request": request,
            "result": result,
            "context": "PYTHON OUTPUT:\n--------------------------------\n" + result,
        }

    def code_execute_host(self, ctx: CtxItem, item: dict, request: dict, all: bool = False) -> dict:
        """
        Execute code on host machine

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :param all: execute all
        :return: response dict
        """
        # write code to file
        data = item["params"]['code']
        if not all:
            path = self.plugin.window.controller.interpreter.file_current
            if "path" in item["params"]:
                path = item["params"]['path']
            path = self.prepare_path(path, on_host=True)
            msg = "Saving Python file: {}".format(path)
            self.log(msg)
            with open(path, 'w', encoding="utf-8") as file:
                file.write(data)
        else:
            path = self.prepare_path(self.plugin.window.controller.interpreter.file_input, on_host=True)

        self.append_input(data)
        self.send_interpreter_input(data)  # send input to interpreter

        # run code
        cmd = self.plugin.get_option_value('python_cmd_tpl').format(filename=path)
        self.log("Running command: {}".format(cmd))
        try:
            process = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
        except Exception as e:
            self.error(e)
            stdout = None
            stderr = str(e).encode("utf-8")
        result = self.handle_result(stdout, stderr)
        return {
            "request": request,
            "result": result,
            "context": "PYTHON OUTPUT:\n--------------------------------\n" + result,
        }

    def code_execute_sandbox(self, ctx, item: dict, request: dict, all: bool = False) -> dict:
        """
        Execute code in sandbox (docker)

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :param all: execute all
        :return: response dict
        """
        data = item["params"]['code']
        if not all:
            path = self.plugin.window.controller.interpreter.file_current
            if "path" in item["params"]:
                path = item["params"]['path']
            msg = "Saving Python file: {}".format(path)
            self.log(msg, sandbox=True)
            with open(self.prepare_path(path, on_host=True), 'w', encoding="utf-8") as file:
                file.write(data)
        else:
            path = self.plugin.window.controller.interpreter.file_input

        self.append_input(data)
        self.send_interpreter_input(data)  # send input to interpreter

        # run code
        path = self.prepare_path(path, on_host=False)
        msg = "Executing Python code: {}".format(item["params"]['code'])
        self.log(msg, sandbox=True)
        cmd = self.plugin.get_option_value('python_cmd_tpl').format(
            filename=path,
        )
        self.log("Running command: {}".format(cmd), sandbox=True)
        response = self.run_docker(cmd)
        result = self.handle_result_docker(response)
        return {
            "request": request,
            "result": result,
            "context": "PYTHON OUTPUT:\n--------------------------------\n" + result,
        }

    def sys_exec_host(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command on host

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg)
        self.log("Running command: {}".format(item["params"]['command']))
        try:
            process = subprocess.Popen(
                item["params"]['command'],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
        except Exception as e:
            self.error(e)
            stdout = None
            stderr = str(e).encode("utf-8")
        result = self.handle_result(stdout, stderr)
        return {
            "request": request,
            "result": result,
            "context": "SYS OUTPUT:\n--------------------------------\n" + result,
        }

    def sys_exec_sandbox(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command in sandbox (docker)

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg, sandbox=True)
        self.log(
            "Running command: {}".format(item["params"]['command']),
            sandbox=True,
        )
        response = self.run_docker(item["params"]['command'])
        result = self.handle_result_docker(response)
        return {
            "request": request,
            "result": result,
            "context": "SYS OUTPUT:\n--------------------------------\n" + result,
        }

    def append_input(self, data: str):
        """
        Append input to interpreter input file

        :param data: input data
        """
        if not self.plugin.get_option_value("attach_output"):
            return

        content = ""
        path = self.prepare_path(self.plugin.window.controller.interpreter.file_input, on_host=True)
        if os.path.isfile(path):
            with open(path, "r", encoding="utf-8") as f:
                content = f.read()
        with open(path, "a", encoding="utf-8") as f:
            nl = ""
            if content != "":
                nl = "\n"
            if data != "":
                f.write(nl + data)
            else:
                f.write("")

    def is_absolute_path(self, path: str) -> bool:
        """
        Check if path is absolute

        :param path: path to check
        :return: True if absolute
        """
        return os.path.isabs(path)

    def prepare_path(self, path: str, on_host: bool = True) -> str:
        """
        Prepare path

        :param path: path to prepare
        :param on_host: is on host
        :return: prepared path
        """
        if self.is_absolute_path(path):
            return path
        else:
            if not self.is_sandbox() or on_host:
                return os.path.join(
                    self.plugin.window.core.config.get_user_dir('data'),
                    path,
                )
            else:
                return path

    def error(self, err: any):
        """
        Log error message

        :param err: exception or error message
        """
        if self.signals is not None:
            self.signals.error.emit(err)

    def status(self, msg: str):
        """
        Send status message

        :param msg: status message
        """
        if self.signals is not None:
            self.signals.status.emit(msg)

    def debug(self, msg: any):
        """
        Log debug message

        :param msg: message to log
        """
        if self.signals is not None:
            self.signals.debug.emit(msg)

    def log(self, msg, sandbox: bool = False):
        """
        Log message to console

        :param msg: message to log
        :param sandbox: is sandbox mode
        """
        prefix = ''
        if sandbox:
            prefix += '[DOCKER]'
        full_msg = prefix + ' ' + str(msg)

        if self.signals is not None:
            self.signals.log.emit(full_msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.24 00:00:00                  #
# ================================================== #

from pygpt_net.provider.web.base import BaseProvider


class Web:
    def __init__(self, window=None):
        """
        Web access core

        :param window: Window instance
        """
        self.window = window
        self.providers = {
            "search_engine": {},
        }

    def is_registered(self, id: str, type: str = "search_engine") -> bool:
        """
        Check if provider is registered

        :param id: provider id
        :param type: provider type
        :return: True if registered
        """
        if type in self.providers:
            return id in self.providers[type]
        return False

    def get_providers(self, type: str = "search_engine") -> dict:
        """
        Get all providers

        :param type: provider type
        :return: providers dict
        """
        if type in self.providers:
            return self.providers[type]
        return {}

    def get_ids(self, type: str = "search_engine") -> list:
        """
        Get all providers ids

        :param type: provider type
        :return: providers ids list
        """
        if type in self.providers:
            return list(self.providers[type].keys())
        return []

    def get(self, id: str, type: str = "search_engine") -> BaseProvider or None:
        """
        Get provider instance

        :param id: provider id
        :param type: provider type
        :return: provider instance
        """
        if self.is_registered(id, type):
            return self.providers[type][id]
        return None

    def register(self, provider: BaseProvider):
        """
        Register provider

        :param provider: provider instance
        """
        id = provider.id
        type = provider.type
        for t in type:
            if t in self.providers:
                self.providers[t][id] = provider



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.31 04:00:00                  #
# ================================================== #

import json
import os
import uuid

from packaging.version import Version

from pygpt_net.provider.core.attachment.base import BaseProvider
from pygpt_net.item.attachment import AttachmentItem


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.id = "json_file"
        self.type = "attachment"
        self.config_file = 'attachments.json'

    def create_id(self) -> str:
        """
        Create unique uuid

        :return: uuid
        """
        return str(uuid.uuid4())

    def create(self, attachment: AttachmentItem) -> str:
        """
        Create new and return its ID

        :param attachment: AttachmentItem
        :return: attachment ID
        """
        if attachment.id is None or attachment.id == "":
            attachment.id = self.create_id()
        return attachment.id

    def load(self) -> dict:
        """
        Load attachments from file

        :return: dict
        """
        path = os.path.join(self.window.core.config.path, self.config_file)
        items = {}
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = json.load(file)
                    if data == "" or data is None or 'items' not in data:
                        return {}
                    # deserialize
                    for mode in data['items']:
                        items[mode] = {}
                        for id in data['items'][mode]:
                            item = data['items'][mode][id]
                            attachment = AttachmentItem()
                            self.deserialize(item, attachment)
                            items[mode][id] = attachment
        except Exception as e:
            self.window.core.debug.log(e)
            items = {}

        return items

    def save(self, items: dict):
        """
        Save attachments to file
        """
        try:
            # update attachments
            path = os.path.join(self.window.core.config.path, self.config_file)
            data = {}
            ary = {}

            # serialize
            for mode in items:
                ary[mode] = {}
                for id in items[mode]:
                    attachment = items[mode][id]
                    ary[mode][id] = self.serialize(attachment)

            data['__meta__'] = self.window.core.config.append_meta()
            data['items'] = ary
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)

        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving attachments: {}".format(str(e)))

    def remove(self, id: str):
        """
        Delete by id

        :param id: id
        """
        pass

    def truncate(self, mode: str):
        """Delete all"""
        path = os.path.join(self.window.core.config.path, self.config_file)
        data = {'__meta__': self.window.core.config.append_meta(), 'items': {}}
        try:
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.core.debug.log(e)

    def patch(self, version: Version) -> bool:
        """
        Migrate presets to current app version

        :param version: current app version
        :return: True if migrated
        """
        return False

    @staticmethod
    def serialize(attachment: AttachmentItem) -> dict:
        """
        Serialize item to dict

        :return: serialized item
        """
        return {
            'id': attachment.id,
            'name': attachment.name,
            'path': attachment.path,
            'remote': attachment.remote,
            'send': attachment.send
        }

    @staticmethod
    def deserialize(data: dict, attachment: AttachmentItem):
        """
        Deserialize item from dict

        :param data: serialized item
        :param attachment: AttachmentItem
        """
        if 'id' in data:
            attachment.id = data['id']
        if 'name' in data:
            attachment.name = data['name']
        if 'path' in data:
            attachment.path = data['path']
        if 'remote_id' in data:
            attachment.remote = data['remote']
        if 'send' in data:
            attachment.send = data['send']

    def dump(self, item: AttachmentItem) -> str:
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.30 19:00:00                  #
# ================================================== #

import copy

from packaging.version import Version

from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.provider.core.attachment.json_file import JsonFileProvider


class Attachments:
    def __init__(self, window=None):
        """
        Attachments core

        :param window: Window instance
        """
        self.window = window
        self.provider = JsonFileProvider(window)
        self.items = {}
        self.current = None

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def select(self, mode: str, id: str):
        """
        Select attachment by uuid

        :param mode: mode
        :param id: id
        """
        if mode not in self.items:
            self.items[mode] = {}

        if id in self.items[mode]:
            self.current = id

    def count(self, mode: str) -> int:
        """
        Count attachments in mode

        :param mode: mode
        :return: attachments count
        """
        if mode not in self.items:
            self.items[mode] = {}

        return len(self.items[mode])

    def get_ids(self, mode: str) -> list:
        """
        Get items IDs in mode

        :param mode: mode
        :return: items UUIDs
        """
        if mode not in self.items:
            self.items[mode] = {}

        return list(self.items[mode].keys())

    def get_id_by_idx(self, mode: str, idx: int) -> str or None:
        """
        Get ID by index in mode

        :param mode: mode
        :param idx: index
        :return: file ID
        """
        i = 0
        for id in self.get_ids(mode):
            if i == idx:
                return id
            i += 1

    def get_by_id(self, mode: str, id: str) -> AttachmentItem or None:
        """
        Return attachment by ID in mode

        :param mode: mode
        :param id: file id
        :return: AttachmentItem
        """
        if mode not in self.items:
            self.items[mode] = {}

        if id in self.items[mode]:
            return self.items[mode][id]

    def get_by_idx(self, mode: str, idx: int) -> AttachmentItem or None:
        """
        Return item by index in mode

        :param mode: mode
        :param idx: item index
        :return: AttachmentItem or None
        """
        id = self.get_id_by_idx(mode, idx)
        if id is not None:
            return self.items[mode][id]

    def get_all(self, mode: str) -> dict:
        """
        Return all items in mode

        :param mode: mode
        :return: attachments items dict
        """
        if mode not in self.items:
            self.items[mode] = {}

        return self.items[mode]

    def new(
            self,
            mode: str,
            name: str = None,
            path: str = None,
            auto_save: bool = True
    ) -> AttachmentItem:
        """
        Create new attachment

        :param mode: mode
        :param name: name
        :param path: path
        :param auto_save: auto_save
        :return: AttachmentItem
        """
        # make local copy of external attachment if enabled
        if self.window.core.config.get("upload.store"):
            if not self.window.core.filesystem.in_work_dir(path):
                path = self.window.core.filesystem.store_upload(path)

        attachment = self.create()
        attachment.name = name
        attachment.path = path

        if mode not in self.items:
            self.items[mode] = {}

        self.items[mode][attachment.id] = attachment
        self.current = attachment.id

        if auto_save:
            self.save()

        return attachment

    def build(self) -> AttachmentItem:
        """
        Build attachment

        :return: AttachmentItem
        """
        attachment = AttachmentItem()
        attachment.name = None
        attachment.path = None
        return attachment

    def create(self) -> AttachmentItem:
        """
        Create attachment item

        :return: AttachmentItem
        """
        attachment = self.build()
        id = self.provider.create(attachment)
        attachment.id = id
        return attachment

    def add(self, mode: str, item: AttachmentItem):
        """
        Add item to attachments

        :param mode: mode
        :param item: item to add to attachments
        """
        if mode not in self.items:
            self.items[mode] = {}

        id = item.id
        self.items[mode][id] = item  # add item to attachments

        # save attachments
        self.save()

    def has(self, mode: str) -> bool:
        """
        Check id mode has attachments

        :param mode: mode
        :return: True if exists
        """
        if mode not in self.items:
            self.items[mode] = {}

        return len(self.items[mode]) > 0

    def delete(self, mode: str, id: str, remove_local: bool = False):
        """
        Delete attachment by file_id

        :param mode: mode
        :param id: file id
        :param remove_local: remove local copy
        """
        if mode not in self.items:
            self.items[mode] = {}

        if id in self.items[mode]:
            if remove_local:
                self.window.core.filesystem.remove_upload(
                    self.items[mode][id].path,
                )
            del self.items[mode][id]
            self.save()

    def delete_all(self, mode: str, remove_local: bool = False):
        """
        Delete all attachments

        :param mode: mode
        :param remove_local: remove local copy
        """
        self.clear(
            mode,
            remove_local=remove_local,
        )
        self.provider.truncate(mode)

    def clear(self, mode: str, remove_local: bool = False):
        """
        Clear all attachments in mode

        :param mode: mode
        :param remove_local: remove local copy
        """
        if mode not in self.items:
            self.items[mode] = {}
        if remove_local:
            for id in self.items[mode]:
                self.window.core.filesystem.remove_upload(
                    self.items[mode][id].path,
                )
        self.items[mode] = {}

    def clear_all(self):
        """
        Clear all attachments
        """
        self.items = {}

    def replace_id(
            self,
            mode: str,
            tmp_id: str,
            attachment: AttachmentItem
    ):
        """
        Replace temporary id with real one

        :param mode: mode
        :param tmp_id: temporary id
        :param attachment: attachment
        """
        if mode not in self.items:
            self.items[mode] = {}

        if tmp_id in self.items[mode]:
            self.items[mode][attachment.id] = self.items[mode][tmp_id]
            del self.items[mode][tmp_id]
            self.save()

    def rename_file(self, mode: str, id: str, name: str):
        """
        Update name

        :param mode: mode
        :param id: file id
        :param name: new name
        """
        data = self.get_by_id(mode, id)
        data.name = name
        self.save()

    def make_json_list(self, attachments: dict) -> dict:
        """
        Make json list

        :param attachments: attachments
        :return: json list
        """
        result = {}
        for id in attachments:
            attachment = attachments[id]
            result[id] = {
                'name': attachment.name,
                'path': attachment.path,
            }
        return result

    def from_files(self, mode: str, files: dict):
        """
        Load current from assistant files

        :param mode: mode
        :param files: files dict
        """
        self.clear(mode)
        for id in files:
            file = files[id]
            item = AttachmentItem()
            item.name = id

            if 'name' in file \
                    and file['name'] is not None and file['name'] != "":
                item.name = file['name']
            if 'path' in file \
                    and file['path'] is not None and file['path'] != "":
                item.path = file['path']

            item.id = id
            item.remote = id
            item.send = True
            self.add(mode, item)

    def from_attachments(self, mode: str, attachments: dict):
        """
        Load current from attachments

        :param mode: mode
        :param attachments: attachments
        """
        self.clear(mode)
        for id in attachments:
            attachment = attachments[id]
            self.add(mode, attachment)

    def load(self):
        """Load attachments"""
        self.items = self.provider.load()
        # replace workdir placeholder with current workdir
        for mode in self.items:
            for id in self.items[mode]:
                attachment = self.items[mode][id]
                if attachment.path is not None:
                    attachment.path = self.window.core.filesystem.to_workdir(
                        attachment.path,
                    )

    def save(self):
        """Save attachments"""
        # replace current workdir with placeholder
        data = copy.deepcopy(self.items)  # copy to avoid changing original data
        for mode in data:
            for id in data[mode]:
                attachment = data[mode][id]
                if attachment.path is not None:
                    attachment.path = self.window.core.filesystem.make_local(
                        attachment.path,
                    )
        self.provider.save(data)
